from typing import Dict, Optional, Tuple, Set, List
import math
from .constants import CROSS_MATCH_ADJUSTMENT


def calculate_score(
    gap_ext: float,
    gap_init: float,
    seq1: str,
    seq2: str,
    prev_char_seq1: str,
    prev_char_seq2: str,
    sub_matrix: Dict[str, int],
    char_complexity_adjustment: Dict[str, float],
) -> float:
    """
    Calculate the score for an alignment between a subfamily and a target/chromsome sequence.
    Scores are calculated based on input sub_matrix, gap_ext, and gap_init.

    prev_char_seq1, prev_char_seq2 are the single nucleotides in the alignment before the chunk - if
    chunk starts with '-' these tell us to use gap_init or gap_ext as the penalty

    input:
    gap_ext: penalty to extend a gap
    gap_init: penalty to start a gap
    seq1: sequence chunk from alignment
    seq2: sequence chunk from alignment
    prev_char_seq1: character before alignment - tells if should use gap_ext or gap_init
    prev_char_seq2: character before alignment - tells if should use gap_ext or gap_init
    sub_matrix: input substitution matrix - dict that maps 2 chars being aligned to the score

    output:
    alignment score

    >>> sub_mat = {"AA":1, "AT":-1, "TA":-1, "TT":1}
    >>> char_adjustments = {'T': 0.0, 'C': 0.0, 'A': 0.0, 'G': 0.0, '-': 0.0}
    >>> calculate_score(-5, -25, "AT", "AT", "", "", sub_mat, char_adjustments)
    2.0
    >>> calculate_score(-5, -25, "-T", "AT", "A", "A", sub_mat, char_adjustments)
    -24.0
    >>> calculate_score(-5, -25, "-T", "AT", "-", "", sub_mat, char_adjustments)
    -4.0
    >>> char_adjustments = {'T': -0.4, 'C': 0.0, 'A': -0.2, 'G': 0.0, '-': 0.0}
    >>> calculate_score(-5, -25, "-T", "AT", "A", "A", sub_mat, char_adjustments)
    -24.4
    >>> calculate_score(-5, -25, "-T", "AT", "-", "", sub_mat, char_adjustments)
    -4.4
    """
    chunk_score: float = 0.0

    # deals with the first character of a segment being a gap character - have to look at last
    # segment to see if this is a gap init or ext
    if seq1[0] == "-":
        if prev_char_seq1 == "-":
            chunk_score += gap_ext
        else:
            chunk_score += gap_init
    elif seq2[0] == "-":
        if prev_char_seq2 == "-":
            chunk_score += gap_ext
        else:
            chunk_score += gap_init
    else:
        chunk_score += sub_matrix[seq1[0] + seq2[0]]
        chunk_score += char_complexity_adjustment[seq2[0]]

    for j in range(1, len(seq1)):
        seq1_char: str = seq1[j]
        seq2_char: str = seq2[j]

        if seq1_char == "-":
            if seq1[j - 1] == "-":
                chunk_score += gap_ext
            else:
                chunk_score += gap_init
        elif seq2_char == "-":
            if seq2[j - 1] == "-":
                chunk_score += gap_ext
            else:
                chunk_score += gap_init
        else:
            chunk_score += sub_matrix[seq1_char + seq2_char]
            chunk_score += char_complexity_adjustment[seq2_char]
    return chunk_score


def calc_target_char_counts(
    query_seq: str,
    target_seq: str,
    target_chars: List[str],
) -> Tuple[Dict[str, int], Set[str]]:
    """
    Finds the char counts in the target sequence which will be used in the calculation
    to determine a character's contribution to the complexity adjusted score.

    input:
    query_seq: query sequence from alignment
    target_seq: target sequence from alignment
    target_chars: chars in the target sequence that will contribute to
    the complexity adjusted score

    output:
    target_char_counts: dictionary of the target chars to their count
    in the target sequence that will contribute to the adjusted score
    other_chars: other chars found in the target sequence that do not contribute
    to the complexity adjustment (gaps, padding)
    """
    target_char_counts: Dict[str, int] = {char: 0 for char in target_chars}
    other_chars = set()
    for q_base, t_base in zip(query_seq, target_seq):
        if t_base not in target_chars:
            other_chars.add(t_base)
            continue
        if q_base == "-":
            continue
        target_char_counts[t_base] += 1
    return target_char_counts, other_chars


def calculate_complexity_adjusted_score(
    char_background_freqs: Optional[Dict[str, float]],
    query_seq: str,
    target_seq: str,
    lamb: float,
) -> Dict[str, float]:
    r"""
    Calculates the per position contribution to the complexity adjusted score
    of the alignment for each valid char in the target sequence.
    This calculation is from the cross_match complexity adjustment functions:

    .. math::
        t_{i} = \text{count of $i^{th}$ character}

    .. math::
        p_{i} = \text{background frequency of $i^{th}$ character}

    .. math::
        t_{factor} = \sum_{i=0}^{N} t_{i}ln(t_{i}) - \left[ \left( \sum_{i=0}^{N} t_{i} \right) ln(\sum_{i=0}^{N} t_{i}) \right]

    .. math::
        t_{sum} = \left[ \sum_{i=0}^{N} t_{i} ln(p_{i}) \right] - t_{factor}

    .. math::
        s_{adj} = s_{raw} + (\frac{t_{sum}}{\lambda}) + 0.999

    input:
    char_background_freqs: background frequencies of the chars from the scoring matrix
    query_seq: query sequence from alignment
    target_seq: target sequence from alignment
    lamb: lambda value used from the scoring matrix

    output:
    char_complexity_adjustments: a char's per position contribution to the
    complexity adjusted score

    >>> query = "TCAGACTGTTCA-----ACTCACCTGGCAGCCACTTCCAGA"
    >>> target = "TCAGACTGTTCATGAGTGCTCACCTGGTAGAGG-----AAA"
    >>> lamb = 0.1227
    >>> calculate_complexity_adjusted_score(None, query, target, lamb)
    {'T': 0.0, 'C': 0.0, 'A': 0.0, 'G': 0.0, '-': 0.0}

    >>> char_freqs = {'A': 0.295, 'G': 0.205, 'C': 0.205, 'T': 0.295}
    >>> calculate_complexity_adjusted_score(char_freqs, query, target, lamb)
    {'A': -0.08342236356495786, 'G': -0.11790190327726593, 'C': -0.11790190327726593, 'T': -0.08342236356495786, '-': 0.0}
    """
    char_complexity_adjustments: Dict[str, float] = {}
    if char_background_freqs is None:
        # set the complexity adjustment to zero for every char
        for char in target_seq:
            char_complexity_adjustments[char] = 0.0
        return char_complexity_adjustments

    t_factor = 0.0
    t_sum = 0.0
    t_counts: int = 0
    target_chars = list(char_background_freqs.keys())
    target_char_counts, other_chars = calc_target_char_counts(
        query_seq, target_seq, target_chars
    )
    total_chars = sum(target_char_counts.values())

    if total_chars > 0:
        char_complexity_adjustments = {char: 0 for char in target_chars}
        for char, freq in char_background_freqs.items():
            count = target_char_counts[char]
            if count > 0 and 0 < freq < 1:
                t_factor += count * math.log(count)
                t_sum += count * math.log(freq)
                t_counts += count
                char_complexity_adjustments[char] += count * math.log(freq)

        # char percent contribution to t_sum
        for c, v in char_complexity_adjustments.items():
            char_complexity_adjustments[c] /= t_sum

        if t_counts != 0:
            t_factor -= t_counts * math.log(t_counts)
        t_sum -= t_factor

        # char value contribution to t_sum
        for c, v in char_complexity_adjustments.items():
            char_complexity_adjustments[c] *= t_sum

        # per position char value contribution to score adjustment
        for c, v in char_complexity_adjustments.items():
            if v != 0:
                char_complexity_adjustments[c] /= target_char_counts[c]
                char_complexity_adjustments[c] /= lamb
                char_complexity_adjustments[c] += (
                    CROSS_MATCH_ADJUSTMENT / total_chars
                )

    # to avoid any key errors
    for char in other_chars:
        char_complexity_adjustments[char] = 0.0

    return char_complexity_adjustments



from typing import List, Tuple


def edges(starts: List[int], stops: List[int]) -> Tuple[int, int]:
    """
    Find and return the min start and max stop positions for the entire region
    included in the alignments.

    Inputs:

    starts - start positions on the target sequence from the input alignment
    stops - stop positions on the target sequence from the input alignment

    Outputs:

    minimum and maximum start and stop positions on the chromosome/target
    sequences for whole alignment

    >>> b, e = edges(starts=[0, 1, 4, 7], stops=[0, 3, 10, 9])
    >>> b
    1
    >>> e
    10
    """
    min_start: int = min(starts[1:])
    max_stop: int = max(stops[1:])

    return min_start, max_stop



from typing import Dict


def sub_repeat_scores(
    start: int, end: int, repeat_scores: Dict[int, float]
) -> float:
    """
    Sums tandem repeat scores from columns start to end.

    input:
    start: column index to start sum
    end: column index to end sum
    repeat_scores: hash implementation of 1d array that maps column index
    to tandem repeat score

    output:
    summ: sum of the tandem repeat scores
    """

    summ: float = 0.0
    curr_index = start
    while curr_index < end:
        if curr_index in repeat_scores:
            summ += repeat_scores[curr_index]
        curr_index += 1
    return summ



from typing import Dict, List, Tuple

from .matrices import ConsensusMatrixContainer
from .performance import timeit


@timeit()
def fill_consensus_position_matrix(
    row_count: int,
    column_count: int,
    start_all: int,
    subfams: List[str],
    chroms: List[str],
    starts: List[int],
    stops: List[int],
    consensus_starts: List[int],
    strands: List[str],
) -> ConsensusMatrixContainer:
    """
    Fills matrix that holds the consensus position for each subfam at that
    position in the alignment. Walks along the alignments one nucleotide at a time adding
    the consensus position to the matrix.

    At same time, fills ActiveCells.

    input:

    column_count: number of columns in alignment matrix - will be same number of
    columns in consensus_matrix
    row_count: number of rows in matrices
    start_all: min start position on chromosome/target sequences for whole alignment
    subfams: actual subfamily/consensus sequences from alignment
    chroms: actual target/chromosome sequences from alignment
    starts: start positions for all competing alignments (on target)
    stops: stop positions for all competing alignments (on target)
    consensus_starts: where alignment starts in the subfam/consensus sequence
    strands: what strand each of the alignments are on - reverse strand will count down instead of up

    output:

    ConsensusMatrixContainer

    >>> subs = ["", ".AA", "TT-"]
    >>> chrs = ["", ".AA", "TTT"]
    >>> strts = [0, 1, 0]
    >>> stps = [0, 2, 2]
    >>> con_strts = [-1, 0, 10]
    >>> strandss = ["", "+", "-"]
    >>> active, con_mat = fill_consensus_position_matrix(3, 3, 0, subs, chrs, strts, stps, con_strts, strandss)
    >>> con_mat
    {(1, 2): 0, (1, 3): 1, (2, 1): 10, (2, 2): 9, (2, 3): 9, (0, 0): 0, (0, 1): 0, (0, 2): 0}
    >>> active
    {2: [0, 1, 2], 3: [0, 1, 2], 1: [0, 2], 0: [0]}
    """
    active_cells: Dict[int, List[int]] = {}
    consensus_matrix: Dict[Tuple[int, int], int] = {}

    # start at 1 to ignore 'skip state'
    for row_index in range(1, row_count):

        if strands[row_index] == "+":
            consensus_pos = consensus_starts[row_index] - 1
            col_index: int = starts[row_index] - start_all + 1
            seq_index: int = 0

            while col_index < stops[row_index] + 1 - start_all + 1:

                # consensus pos only advances when there is not a gap in the subfam seq
                if subfams[row_index][seq_index] != "-":
                    consensus_pos += 1

                consensus_matrix[row_index, col_index] = consensus_pos

                # matrix position only advances when there is not a gap in the chrom seq
                if chroms[row_index][seq_index] != "-":
                    if col_index in active_cells:
                        active_cells[col_index].append(row_index)
                    else:
                        active_cells[col_index] = [0, row_index]
                    col_index += 1

                seq_index += 1

        else:  # reverse strand
            consensus_pos2 = consensus_starts[row_index] + 1
            col_index2: int = starts[row_index] - start_all + 1
            seq_index2: int = 0

            while col_index2 < stops[row_index] + 1 - start_all + 1:

                if subfams[row_index][seq_index2] != "-":
                    consensus_pos2 -= 1
                consensus_matrix[row_index, col_index2] = consensus_pos2

                if chroms[row_index][seq_index2] != "-":
                    if col_index2 in active_cells:
                        active_cells[col_index2].append(row_index)
                    else:
                        active_cells[col_index2] = [0, row_index]
                    col_index2 += 1

                seq_index2 += 1

    for i in range(column_count):
        consensus_matrix[0, i] = 0
        if i not in active_cells:
            active_cells[i] = [0]

    return ConsensusMatrixContainer(active_cells, consensus_matrix)



from typing import List, NamedTuple


class Alignment(NamedTuple):
    """
    A container to hold data related to a single alignment.

    >>> a = Alignment("", "ch", 25, 100, 0, 0, 0, 0, 0, [], "", 0, "", 0, 0)
    >>> a.chrom_name
    'ch'
    >>> a.chrom_start
    25
    >>> a.chrom_stop
    100
    >>> a.chrom_length
    76
    """

    subfamily: str
    chrom_name: str
    chrom_start: int
    chrom_stop: int
    score: int
    start: int
    stop: int
    consensus_start: int
    consensus_stop: int
    sequences: List[str]
    strand: str
    flank: int
    sub_matrix_name: str
    gap_init: float
    gap_ext: float

    @property
    def chrom_length(self) -> int:
        return self.chrom_stop - self.chrom_start + 1

    @property
    def sequence(self) -> str:
        return self.sequences[0]

    @sequence.setter
    def sequence(self, value) -> None:
        self.sequences[0] = value

    @property
    def subfamily_sequence(self) -> str:
        return self.sequences[1]

    @subfamily_sequence.setter
    def subfamily_sequence(self, value) -> None:
        self.sequences[1] = value


_skip = Alignment(
    subfamily="skip",
    chrom_name="skip_chrom",
    chrom_start=0,
    chrom_stop=0,
    score=0,
    start=0,
    stop=0,
    consensus_start=0,
    consensus_stop=0,
    sequences=["", ""],
    strand="",
    flank=0,
    sub_matrix_name="",
    gap_init=0,
    gap_ext=0,
)


def get_skip_state():
    """
    Return the skip state singleton.
    """
    return _skip



from polyA.fill_align_matrix import fill_align_matrix
from polyA.edges import edges
from polyA.constants import SKIP_ALIGN_SCORE
from polyA.lambda_provider import ConstantLambdaProvider
from polyA.substitution_matrix import load_substitution_matrices, SubMatrix


def test_fill_align_matrix():
    chrom_seqs = [
        "",
        "TCAGACTGTTCATGAGTGCTCACCTGGTAGAGG-----AAAGACCCTG-AAC----GTCCAGAGCTTTCCCCTGACTGGCCACGTGCGGACTCTTTGTGGCT---CTGCTGA................",
        "CCATCTTTAAGAATGGAAGATGGACATGAG-AAACTTTTCT--CTA-----CCATCTACATAAAACAAATGA--ATAGAGTTTTCCTGACTTAA---ATAAAATTGCCCATTTTTTGGCCTTGTGCC-CACTTA------CAGCTGGACTTAGTAGTCTCACA--GACCTCAGTGTGCTCTAGGTAAGATCATCCCTCTTGCTGATCTTGCCATTGCTAGCAACAAGAATTTAGAAAGGCTGACT................",
        "TTAAATGACTTTTAATGAAGGGGTTT-TCGATTGAATTATTTACTAATTAAACGA--TTCTGTGTTAAGAGACTTTCGGTCAATCGGTATACTGTTCTATGAGTTACT-TTCAAG-----TAA--GTTAAAAAA----TAAACAA--AAAATTTGTAAAGTACTTCAAATATTATAAAG--TAACATATCCTACCTAAATTATCGAGAAGTTATAAATTTCTCTTTCCAAAA-AGACTGACCCAAAGTAAAAC................",
    ]
    subfam_seqs = [
        "",
        "TCAGACTGTTCA-----ACTCACCTGGCAGCCACTTCCAGAGCCCCTGGAACTCTAGCCCAAGGCTCTC---TGACTGACCCCTTCTGAGATCTTCTTGGCTTAGCAGCTGA................",
        "CCAGCCTTAGGACTGCCAGAT----ATAACTAAGCCTTTCTTTCTATATGTCCATGAACG-AAAAGGAATGGCTATAGGGGT--CCTGACTCAAGTCATAGGAT----CATGGCATCGCCCGGTGGCATACTTATCCAAGCAACTGGACTCCGTGGCNCTAGGATGGCCTC--CTTGCCTTAGG---G--CA-----CTAGCTGNCACTGCC-CTACTGGCA-CAAGAAGCTAACAAA-CTGACT................",
        "TTAAGCGGCTTTCGGTTAAGCGGCTTACCGGTTAAGCGGCTT-TTAGTTAAGCGGCTTTCTG-GTTAAGCGGCTTTCGGTTAAGCGGCTTTTGGTTAAGCGGCTTACTGGTTAAGCGGTTTAATGGTTAAATGGTTTTTGAACAAAGAAAATTGATAAA-TATTTCAAATATCGTTAAGCATAACCTACCCTACCAAAATTGTCGGAAACTCCTAAA-AACGCTT--CAAAATCAACTGGCCAAAANTAAAAC................",
    ]
    alignment_start_positions = [6409, 6410, 6450, 6463]
    alignment_stop_positions = [6696, 6508, 6672, 6695]
    gap_inits = [0, -25.0, -25.0, -25.0]
    gap_exts = [0, -5.0, -5.0, -5.0]
    start_all, stop_all = edges(
        alignment_start_positions, alignment_stop_positions
    )
    column_count = stop_all - start_all + 1 + 2

    _lambda_provider = ConstantLambdaProvider(0.1227)
    with open(
        f"fixtures/ultra_test_files/ex13.fa.cm.matrix", "r"
    ) as _sub_matrices_file:
        sub_matrices = load_substitution_matrices(
            _sub_matrices_file, _lambda_provider, False
        )
    lambda_values = [0.0]
    alignment_substitution_matrices = [SubMatrix("skip", 0.0)]

    for seq in chrom_seqs:
        lambda_values.append(sub_matrices["matrix1"].lamb)
        alignment_substitution_matrices.append(sub_matrices["matrix1"])

    align_matrix = fill_align_matrix(
        lambda_values,
        column_count,
        start_all,
        31,
        gap_inits,
        gap_exts,
        SKIP_ALIGN_SCORE,
        subfam_seqs,
        chrom_seqs,
        alignment_start_positions,
        alignment_stop_positions,
        [sm.scores for sm in alignment_substitution_matrices],
        [sm.background_freqs for sm in alignment_substitution_matrices],
    )
    # check values of align matrix
    # assert certain values exists or don't exist
    assert (1, 0) not in align_matrix
    assert (1, 100) not in align_matrix
    assert align_matrix[1, 1] == 15.690262500000001
    assert align_matrix[1, 99] == 12.59975625

    assert (2, 40) not in align_matrix
    assert (2, 264) not in align_matrix
    assert align_matrix[2, 41] == 13.075218750000001
    assert align_matrix[2, 263] == 4.99235625

    assert (3, 53) not in align_matrix
    assert (3, 287) not in align_matrix
    assert align_matrix[3, 54] == 17.11665
    assert align_matrix[3, 286] == 22.822200000000002

    assert align_matrix[0, 0] == 10.0
    assert align_matrix[0, 287] == 10.0


def test_fill_align_matrix_complexity_adjusted():
    subfam_seqs = [
        "",
        "TCAGACTGTTCA-----ACTCACCTGGCAGCCACTTCCAGAGCCCCTGGAACTCTAGCCCAAGGCTCTC---TGACTGACCCCTTCTGAGATCTTCTTGGCTTAGCAGCTGA................",
        "CCAGCCTTAGGACTGCCAGAT----ATAACTAAGCCTTTCTTTCTATATGTCCATGAACG-AAAAGGAATGGCTATAGGGGT--CCTGACTCAAGTCATAGGAT----CATGGCATCGCCCGGTGGCATACTTATCCAAGCAACTGGACTCCGTGGCNCTAGGATGGCCTC--CTTGCCTTAGG---G--CA-----CTAGCTGNCACTGCC-CTACTGGCA-CAAGAAGCTAACAAA-CTGACT................",
        "TTAAGCGGCTTTCGGTTAAGCGGCTTACCGGTTAAGCGGCTT-TTAGTTAAGCGGCTTTCTG-GTTAAGCGGCTTTCGGTTAAGCGGCTTTTGGTTAAGCGGCTTACTGGTTAAGCGGTTTAATGGTTAAATGGTTTTTGAACAAAGAAAATTGATAAA-TATTTCAAATATCGTTAAGCATAACCTACCCTACCAAAATTGTCGGAAACTCCTAAA-AACGCTT--CAAAATCAACTGGCCAAAANTAAAAC................",
    ]
    chrom_seqs = [
        "",
        "TCAGACTGTTCATGAGTGCTCACCTGGTAGAGG-----AAAGACCCTG-AAC----GTCCAGAGCTTTCCCCTGACTGGCCACGTGCGGACTCTTTGTGGCT---CTGCTGA................",
        "CCATCTTTAAGAATGGAAGATGGACATGAG-AAACTTTTCT--CTA-----CCATCTACATAAAACAAATGA--ATAGAGTTTTCCTGACTTAA---ATAAAATTGCCCATTTTTTGGCCTTGTGCC-CACTTA------CAGCTGGACTTAGTAGTCTCACA--GACCTCAGTGTGCTCTAGGTAAGATCATCCCTCTTGCTGATCTTGCCATTGCTAGCAACAAGAATTTAGAAAGGCTGACT................",
        "TTAAATGACTTTTAATGAAGGGGTTT-TCGATTGAATTATTTACTAATTAAACGA--TTCTGTGTTAAGAGACTTTCGGTCAATCGGTATACTGTTCTATGAGTTACT-TTCAAG-----TAA--GTTAAAAAA----TAAACAA--AAAATTTGTAAAGTACTTCAAATATTATAAAG--TAACATATCCTACCTAAATTATCGAGAAGTTATAAATTTCTCTTTCCAAAA-AGACTGACCCAAAGTAAAAC................",
    ]
    alignment_start_positions = [6409, 6410, 6450, 6463]
    alignment_stop_positions = [6696, 6508, 6672, 6695]
    gap_inits = [0, -25.0, -25.0, -25.0]
    gap_exts = [0, -5.0, -5.0, -5.0]
    start_all, stop_all = edges(
        alignment_start_positions, alignment_stop_positions
    )
    column_count = stop_all - start_all + 1 + 2

    _lambda_provider = ConstantLambdaProvider(0.1227)
    with open(
        f"fixtures/ultra_test_files/ex13.fa.cm.matrix", "r"
    ) as _sub_matrices_file:
        sub_matrices = load_substitution_matrices(
            _sub_matrices_file, _lambda_provider, True
        )
    lambda_values = [0.0]
    alignment_substitution_matrices = [SubMatrix("skip", 0.0)]

    for seq in chrom_seqs:
        lambda_values.append(sub_matrices["matrix1"].lamb)
        alignment_substitution_matrices.append(sub_matrices["matrix1"])

    align_matrix = fill_align_matrix(
        lambda_values,
        column_count,
        start_all,
        31,
        gap_inits,
        gap_exts,
        SKIP_ALIGN_SCORE,
        subfam_seqs,
        chrom_seqs,
        alignment_start_positions,
        alignment_stop_positions,
        [sm.scores for sm in alignment_substitution_matrices],
        [sm.background_freqs for sm in alignment_substitution_matrices],
    )

    assert (1, 0) not in align_matrix
    assert (1, 100) not in align_matrix
    assert align_matrix[1, 1] == 14.762038466561917
    assert align_matrix[1, 99] == 11.333849987455658

    assert (2, 40) not in align_matrix
    assert (2, 264) not in align_matrix
    assert align_matrix[2, 41] == 12.98841727761375
    assert align_matrix[2, 263] == 4.912091167445111

    assert (3, 53) not in align_matrix
    assert (3, 287) not in align_matrix
    assert align_matrix[3, 54] == 15.686524697939982
    assert align_matrix[3, 286] == 21.288106662982315

    assert align_matrix[0, 0] == 10.0
    assert align_matrix[0, 287] == 10.0



from typing import List


def pad_sequences(
    chunk_size: int,
    subfam_seqs: List[str],
    chrom_seqs: List[str],
    padding_char: str = ".",
) -> None:
    """
    Right-pad sequences with `(chunk_size - 1) / 2` copies of ".".

    The first sequence is ignored since that is the skip state.

    The `padding_char` parameter exists because the doctest doesn't "see" the
    periods, so the test will pass no matter how many are added to each
    sequence.

    Inputs:

    chunk_size: size (in nucleotides) of each segment
    subfam_seqs: actual subfamily / consensus sequences from alignment
    chrom_seqs: actual target / chromosome sequences from alignment
    padding_char: for testing, specifies the character to use for padding

    Side effects:

    Updates subfam_seqs and chrom_seqs with padded sequences

    >>> s_seq = ['', 'a', 'aaa']
    >>> c_seq = ['', 'a', 't-t']
    >>> pad_sequences(31, s_seq, c_seq, padding_char="-")
    >>> s_seq
    ['', 'a----------------', 'aaa----------------']
    >>> c_seq
    ['', 'a----------------', 't-t----------------']
    """
    padding = padding_char * ((chunk_size + 1) // 2)

    for i in range(1, len(subfam_seqs)):
        subfam_seqs[i] = subfam_seqs[i] + padding
        chrom_seqs[i] = chrom_seqs[i] + padding



from typing import Dict, List
import re


def confidence_cm(
    region: List[float],
    subfam_counts: Dict[str, float],
    subfams: List[str],
    subfam_rows: List[int],
    repeats: int,
    node_confidence_bool: bool,
) -> List[float]:
    """
    Computes confidence values for competing annotations using alignment and tandem
    repeat scores. Loops through the array once to find sum of 2^every_hit_score in
    region, then loops back through to calculate confidence. Converts the alignment
    score to account for lambda before summing.

    If command line option for subfam_counts, this is included in confidence math.

    input:
    region: list of scores for competing annotations
    subfam_counts: dict that maps subfam name to it's count info
    subfams: list of subfam names
    subfam_rows: list of subfamily rows that correspond to the the subfams of the region scores
    repeats: number of tandem repeat scores found at the end of the region
    node_confidence_bool: if False (0) - confidence for filling DP matrices, if True - confidence for nodes

    output:
    confidence_list: list of confidence values for competing annotations, each input alignment
    and tandem repeat score will have one output confidence score

    >>> counts = {"s1": .33, "s2": .33, "s3": .33}
    >>> subs = ["s1", "s2", "s3"]
    >>> conf = confidence_cm([2, 1, 1], counts, subs, [0, 1, 2], 0, False)
    >>> f"{conf[0]:.2f}"
    '0.50'
    >>> f"{conf[1]:.2f}"
    '0.25'
    >>> f"{conf[2]:.2f}"
    '0.25'

    >>> conf = confidence_cm([0, 100, 100], 0, subs, [0, 1, 2], 0, False)
    >>> f"{conf[0]:.2f}"
    '0.01'

    >>> conf = confidence_cm([0, 100, 100], 0, subs, [0, 1, 2], 0, True)
    >>> f"{conf[0]:.2f}"
    '0.00'

    >>> counts = {"s1": .31, "s2": .31, "s3": .31, "Tandem Repeat": .06}
    >>> subs = ["s1", "s2", "s3", "Tandem Repeat"]
    >>> conf = confidence_cm([2, 1, 0.7], counts, subs, [0, 1, 3], 1, False)
    >>> f"{conf[0]:.2f}"
    '0.65'
    >>> f"{conf[1]:.2f}"
    '0.32'
    >>> f"{conf[2]:.2f}"
    '0.03'
    """

    confidence_list: List[float] = []
    score_total: int = 0

    # if command line option to include subfam_counts
    if subfam_counts:
        # alignment scores
        for index in range(len(region) - repeats):
            subfam: str = subfams[subfam_rows[index]]
            m = re.search(r"(.+?)#.+", subfams[subfam_rows[index]])
            if m:
                subfam = m.group(1)
            converted_score = (2 ** int(region[index])) * subfam_counts[subfam]
            confidence_list.append(converted_score)
            score_total += converted_score
        # TR scores
        for index in range(len(region) - repeats, len(region)):
            subfam = subfams[subfam_rows[index]]
            m = re.search(r"(.+?)#.+", subfams[subfam_rows[index]])
            if m:
                subfam = m.group(1)
            tr_score = (2 ** int(region[index])) * subfam_counts[subfam]
            confidence_list.append(tr_score)
            score_total += tr_score

    # don't include subfam counts (default)
    else:
        # alignment scores
        for index in range(len(region) - repeats):
            converted_score = 2 ** int(region[index])
            confidence_list.append(converted_score)
            score_total += converted_score
        # TR scores
        for index in range(len(region) - repeats, len(region)):
            tr_score = 2 ** int(region[index])
            confidence_list.append(tr_score)
            score_total += tr_score

    for index in range(len(region)):
        confidence_list[index] = confidence_list[index] / score_total

    # if skip state confidence is < 1 %, increase it to 1 % and normalize all
    # others do not do this when computing node confidence (skip state is not
    # used)
    if confidence_list[0] < 0.01 and not node_confidence_bool:
        summ = 0.0
        for i in range(1, len(confidence_list)):
            summ += confidence_list[i]

        confidence_list[0] = 0.01
        for i in range(1, len(confidence_list)):
            confidence_list[i] = confidence_list[i] * 0.99 / summ

    return confidence_list


def confidence_only(
    region: List[int],
    lambs: List[float],
) -> List[float]:
    """
    Computes confidence values for competing annotations using alignment scores. Loops
    through the array once to find sum of 2^every_hit_score in
    region, then loops back through to calculate confidence. Converts the alignment
    score to account for lambda before summing.

    input:
    lambdaa: lambda for score matrix used
    region: list of scores for competing annotations
    subfams: list of subfam names

    output:
    confidence_list: list of confidence values for competing annotations

    >>> reg = [100, 55, 1]
    >>> lambs = [.1227] * 3
    >>> conf = confidence_only(reg, lambs)
    >>> conf
    [0.9843787551069454, 0.015380918048546022, 0.0002403268445085316]
    """

    confidence_list: List[float] = []
    score_total: int = 0

    # alignment scores
    for index in range(len(region)):
        converted_score = 2 ** int(region[index] * lambs[index])
        confidence_list.append(converted_score)
        score_total += converted_score

    for index in range(len(region)):
        confidence_list[index] = confidence_list[index] / score_total

    return confidence_list



from math import inf
from uuid import uuid4
from typing import Dict, List, Tuple

from polyA.performance import timeit


@timeit()
def get_path(
    columns: List[int],
    ids: List[str],
    subfams_collapse: List[str],
    last_column: List[float],
    active_cells_collapse: Dict[int, List[int]],
    origin_matrix: Dict[Tuple[int, int], int],
    same_subfam_change_matrix: Dict[Tuple[int, int], int],
) -> Tuple[List[int], List[str]]:
    """
    back traces through origin matrix to get most probable path through the DP matrix.
    Finds where the path switches to a different row and populates Changes and ChangesPosition.
    Reverses Changes and ChangesPosition because it's a backtrace so they are initially backwards.
    Jumps over removed/empty columns when necessary.

    assigns IDs to each col in matrices (corresponds to a nucleotide position in target/chrom
    sequence) - cols with same ID are part of same initial subfam

    input:
    temp_id: current id number being used - makes it so new ids are unique
    columns: list of non empty columns in matrices
    ids: list of ids for each column in matrices
    changes_orig: original changes from first DP trace
    changes_pos_orig: original changes_position from first DP trace
    columns_orig: original list of non empty columns before any nodes are extracted
    subfams_collapse: subfamily names for the rows
    last_column: last column in probability matrix. Used to find where to start backtrace.
    active_cells_collapse: holds which rows haves values for each column
    origin_matrix: origin matrix
    same_subfam_change_matrix: parallel to origin_matrix, if 1 - came from the same subfam, but
    got a change transition probability

    output:
    temp_id: updated current id number being used after function completes
    changes_position: which columns (positions in target/chrom seq) switch to different subfam
    changes: parallel array to changes_position - what subfam is being switches to
    updates input list ids

    >>> non_cols = [0, 1, 2, 3]
    >>> idss = ["", "", "", ""]
    >>> subs = ["s1", "s2"]
    >>> active_col = {0: [0, 1], 1: [0, 1], 2: [0, 1], 3: [0, 1]}
    >>> last_col = [-100, -10]
    >>> orig_mat = {(0, 0): 0, (1, 0): 1, (0, 1): 0, (1, 1): 0, (0, 2): 0, (1, 2): 0, (0, 3): 0, (1, 3): 1}
    >>> same_sub_mat = {}
    >>> (changes_pos, changess) = get_path(non_cols, idss, subs, last_col, active_col, orig_mat, same_sub_mat)
    >>> changes_pos
    [0, 2, 3]
    >>> changess
    ['s1', 's2']
    >>> idss[0] == idss[1]
    True
    >>> idss[2] == idss[3]
    True
    >>> idss[0] != idss[2]
    True
    """

    maxxx: float = -inf
    max_row_index: int = 0

    changes_position: List[int] = []
    changes: List[str] = []

    # which row to start backtrace
    for i in range(len(active_cells_collapse[columns[-1]])):
        if maxxx < last_column[i]:
            maxxx = last_column[i]
            max_row_index = active_cells_collapse[columns[-1]][i]

    prev_row_index: int = origin_matrix[max_row_index, columns[-1]]

    current_id = uuid4().hex

    ids[columns[-1]] = current_id

    # last column is a skip state pad
    changes_position.append(len(columns) - 1)

    # already added the last col, but this adds the one before cols so still start at last col
    for columns_index in range(len(columns) - 1, 1, -1):

        prev_column: int = columns[columns_index - 1]

        ids[columns[columns_index - 1]] = current_id

        if prev_row_index != origin_matrix[prev_row_index, prev_column]:
            current_id = uuid4().hex
            changes_position.append(columns_index - 1)
            changes.append(subfams_collapse[prev_row_index])
        else:
            if (prev_row_index, prev_column) in same_subfam_change_matrix:
                current_id = uuid4().hex
                changes_position.append(columns_index - 1)
                changes.append(subfams_collapse[prev_row_index])

        prev_row_index = origin_matrix[prev_row_index, prev_column]

    ids[columns[0]] = current_id
    changes_position.append(0)
    changes.append(subfams_collapse[prev_row_index])

    changes.reverse()
    changes_position.reverse()

    return changes_position, changes



from typing import Dict, List, Tuple

from .confidence_cm import confidence_cm
from .performance import timeit
from .substitution_matrix import SubMatrix
from .calculate_score import (
    calculate_score,
    calculate_complexity_adjusted_score,
)
from .sum_repeat_scores import sub_repeat_scores


@timeit()
def fill_node_confidence(
    nodes: int,
    start_all: int,
    gap_inits: List[float],
    gap_exts: List[float],
    columns: List[int],
    starts: List[int],
    stops: List[int],
    changes_position: List[int],
    subfams: List[str],
    subfam_seqs: List[str],
    chrom_seqs: List[str],
    subfam_countss: Dict[str, float],
    sub_matrices: List[SubMatrix],
    repeat_scores: Dict[int, float],
    tr_count: int,
) -> Dict[Tuple[str, int], float]:
    """
    for a particular annotated node, takes all competing alignments and calculates
    the confidence for that node

    first fills matrix with node alignment scores, then reuses matrix for confidence scores
    Using changes_position indentifies boundaries for all nodes, and computes confidence
    values for each node.

    input:
    all input needed for CalcScore() and ConfidenceCM()
    nodes: number of nodes
    changes_pos: node boundaries
    columns: all non empty columns in matrices

    output:
    node_confidence: Hash implementation of sparse 2D matrix that holds confidence values
    for whole nodes. Used during stitching process. Tuple[str, int] is key that maps a subfamily
    and node number to a confidence score.

    >>> non_cols = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> strts = [0, 0, 0]
    >>> stps = [10, 10, 10]
    >>> change_pos = [1, 3, 7, 10]
    >>> names = ["skip", "n1", "n2"]
    >>> s_seqs = ['', 'AAA-TTTT-T', 'TTTTTTTTTT']
    >>> c_seqs = ['', 'TTTTTTTTTT', 'TTTTTTTTTT']
    >>> counts = {"skip": .33, "n1": .33, "n2": .33}
    >>> sub_mat = SubMatrix("", 0.1227)
    >>> sub_mat.scores = {"AA": 10, "AT": -1, "TA": -1, "TT": 10, "..":0}
    >>> sub_mat.background_freqs = None
    >>> sub_mats = [sub_mat] * 3
    >>> rep_scores = {}
    >>> node_conf = fill_node_confidence(3, 0, [0, -25, -25], [0, -5, -5], non_cols, strts, stps, change_pos, names, s_seqs, c_seqs, counts, sub_mats, rep_scores, 0)
    >>> node_conf
    {('n1', 0): 0.19999999999999998, ('n2', 0): 0.7999999999999999, ('n1', 1): 0.058823529411764705, ('n2', 1): 0.9411764705882353, ('n1', 2): 0.1111111111111111, ('n2', 2): 0.8888888888888888}
    >>> s_seqs = ['', 'AAA-T--TT-', 'TTTTTTTTTT']
    >>> c_seqs = ['', 'TTTTTTTTTT', 'TTTTTTTTTT']
    >>> node_conf2 = fill_node_confidence(3, 0, [0, -25, -25], [0, -5, -5], non_cols, strts, stps, change_pos, names, s_seqs, c_seqs, counts, sub_mats, rep_scores, 0)
    >>> node_conf2
    {('n1', 0): 0.19999999999999998, ('n2', 0): 0.7999999999999999, ('n1', 1): 0.001949317738791423, ('n2', 1): 0.9980506822612085, ('n1', 2): 0.1111111111111111, ('n2', 2): 0.8888888888888888}
    """
    node_confidence: Dict[Tuple[str, int], float] = {}
    node_confidence_temp: Dict[Tuple[int, int], float] = {}
    active: Dict[int, List[int]] = {}

    # no need to look at the skip nodes
    node_non_empty: List[int] = []

    # matrix colunms doesn't always equal sequence position because of gaps
    # for each matrix column, compute the gap offset for the sequence
    # sequence position = matrix position + offset
    gap_offset: List[List[int]] = [[] for _ in range(len(subfams))]
    for chrom_index in range(len(chrom_seqs)):
        offset = 0
        for seq_index in range(len(chrom_seqs[chrom_index])):
            if chrom_seqs[chrom_index][seq_index] == "-":
                offset += 1
            else:
                gap_offset[chrom_index].append(offset)

    for node_index in range(nodes):
        begin_node = columns[changes_position[node_index]]
        end_node = columns[changes_position[node_index + 1] - 1]
        range_in_columns = (
            changes_position[node_index + 1] - changes_position[node_index]
        )
        if node_index == nodes - 1:
            # for the last node, include the end column
            end_node += 1

        for subfam_index in range(1, len(subfams) - tr_count):
            subfam_start = starts[subfam_index] - start_all + 1
            subfam_stop = stops[subfam_index] - start_all + 1

            sub_matrix = sub_matrices[subfam_index]
            lamb = sub_matrix.lamb
            gap_init = gap_inits[subfam_index]
            gap_ext = gap_exts[subfam_index]

            if subfam_start <= end_node and subfam_stop >= begin_node:
                # subfam in node, calculate alignment score
                last_prev_subfam = ""
                last_prev_chrom = ""
                alignment_index_start = begin_node - subfam_start

                last_index = 0

                if (
                    0
                    <= alignment_index_start - 1
                    <= len(gap_offset[subfam_index])
                ):
                    chrom_offset = gap_offset[subfam_index][
                        alignment_index_start - 1
                    ]
                    last_prev_subfam = subfam_seqs[subfam_index][
                        alignment_index_start - 1 + chrom_offset
                    ]
                    last_prev_chrom = chrom_seqs[subfam_index][
                        alignment_index_start - 1 + chrom_offset
                    ]

                for j in range(
                    changes_position[node_index],
                    changes_position[node_index] + range_in_columns,
                ):
                    i = columns[j] - begin_node
                    if (
                        0
                        <= alignment_index_start + i
                        < len(gap_offset[subfam_index])
                    ):
                        chrom_offset = gap_offset[subfam_index][
                            alignment_index_start + i
                        ]

                        first_index = alignment_index_start + i + chrom_offset
                        break

                for j in range(
                    changes_position[node_index] + range_in_columns - 1,
                    changes_position[node_index],
                    -1,
                ):
                    i = columns[j] - begin_node
                    if (
                        0
                        <= alignment_index_start + i
                        < len(gap_offset[subfam_index])
                    ):
                        chrom_offset = gap_offset[subfam_index][
                            alignment_index_start + i
                        ]
                        last_index = alignment_index_start + i + chrom_offset
                        break

                # add padding first in case no part of seq is used
                chrom_seq = (
                    "." + chrom_seqs[subfam_index][first_index : last_index + 1]
                )
                subfam_seq = (
                    "."
                    + subfam_seqs[subfam_index][first_index : last_index + 1]
                )

                char_complexity_adjustments = (
                    calculate_complexity_adjusted_score(
                        sub_matrix.background_freqs, subfam_seq, chrom_seq, lamb
                    )
                )

                align_score = lamb * calculate_score(
                    gap_ext,
                    gap_init,
                    subfam_seq,
                    chrom_seq,
                    last_prev_subfam,
                    last_prev_chrom,
                    sub_matrix.scores,
                    char_complexity_adjustments,
                )

                node_confidence_temp[subfam_index, node_index] = align_score

                if node_index in active:
                    active[node_index].append(subfam_index)
                else:
                    active[node_index] = [subfam_index]
                    node_non_empty.append(node_index)

        # TRs
        for subfam_index in range(len(subfams) - tr_count, len(subfams)):
            rep_sum_score: float = 0.0
            tr_start = starts[subfam_index] - start_all + 1
            tr_stop = stops[subfam_index] - start_all + 1

            if not (tr_start > end_node or tr_stop < begin_node):
                rep_sum_score = sub_repeat_scores(
                    begin_node, end_node, repeat_scores
                )

            node_confidence_temp[subfam_index, node_index] = rep_sum_score

            if node_index in active:
                active[node_index].append(subfam_index)
            else:
                active[node_index] = [subfam_index]
                node_non_empty.append(node_index)

    # reuse same matrix and compute confidence scores for the nodes
    for node_index in node_non_empty:
        temp: List[float] = []
        for row_index in active[node_index]:
            temp.append(node_confidence_temp[row_index, node_index])
        confidence_temp: List[float] = confidence_cm(
            temp, subfam_countss, subfams, active[node_index], tr_count, True
        )
        for row_index2 in range(len(confidence_temp)):
            node_confidence_temp[
                active[node_index][row_index2], node_index
            ] = confidence_temp[row_index2]

    # collapse node_confidence down same way supportmatrix is collapsed - all seqs of
    # the same subfam are put in the same row
    for node_index in node_non_empty:
        for row_index in active[node_index]:
            if (subfams[row_index], node_index) in node_confidence:
                node_confidence[
                    subfams[row_index], node_index
                ] += node_confidence_temp[row_index, node_index]
            else:
                node_confidence[
                    subfams[row_index], node_index
                ] = node_confidence_temp[row_index, node_index]

    node_confidence_temp.clear()

    return node_confidence



import re
from typing import Iterable, List, NamedTuple, Optional, TextIO, Tuple

from .exceptions import FileFormatException
from .alignment import Alignment, get_skip_state
from .constants import INFINITE_SHARD_GAP
from .performance import timeit


def _parse_meta_line(line: str) -> Optional[Tuple[str, str]]:
    """
    Attempt to parse one line of a Stockholm file. If the line
    contains metadata, return a tuple of the key and value.
    Otherwise, return ``None``.

    >>> _parse_meta_line("#=GF ID\tSeqMcSeq")
    ('ID', 'SeqMcSeq')
    >>> _parse_meta_line("seqname\tatcg")
    >>> _parse_meta_line("//")
    >>> _parse_meta_line("# stockholm 1.0")
    """
    if line.strip().startswith("#="):
        parts = line.strip().split()

        if len(parts) != 3:
            return None

        return parts[1], parts[2]

    return None


def _parse_alignment_line(line: str) -> Optional[Tuple[str, str]]:
    """
    Attempt to parse one line of a sequence alignment,
    returning a tuple of the sequence name and the sequence
    itself. If the string isn't formatted correctly, return
    ``None``.

    >>> _parse_alignment_line("seqname\tatcg")
    ('seqname', 'atcg')
    >>> _parse_alignment_line("#=GF ID SeqMcSeq")
    >>> _parse_meta_line("//")
    >>> _parse_meta_line("# stockholm 1.0")
    """
    parts = line.strip().split()

    if len(parts) != 2:
        return None

    return parts[0], parts[1]


def _parse_preamble_line(line: str) -> bool:
    """
    Return ``True`` if this line contains a Stockholm format
    preamble ("# Stockholm 1.0") and ``False`` otherwise.
    """
    return line.strip().upper().startswith("# STOCKHOLM")


def _parse_terminator_line(line: str) -> bool:
    """
    Return ``True`` if this line contains a terminator
    ("//") or ``False`` otherwise.

    >>> _parse_terminator_line(" // ")
    True
    >>> _parse_terminator_line(" / / ")
    False
    """
    return line.strip() == "//"


def _parse_chrom_meta(line: str) -> Optional[Tuple[str, int, int]]:
    match = re.search(r"(.+):(\d+)-(\d+)", line.strip())
    if match is None:
        return None

    chrom_name = match.groups()[0]
    chrom_start = int(match.groups()[1])
    chrom_stop = int(match.groups()[2])

    return chrom_name, chrom_start, chrom_stop


@timeit()
def load_alignments(
    file: TextIO, add_skip_state: bool = False
) -> Iterable[Alignment]:
    """
    Load a set of alignments in Stockholm format.
    See below for an example.

    ::

        #=GF ID  MERX#DNA/TcMar-Tigger
        #=GF TR  chr0:0000-0000
        #=GF SC  1153
        #=GF ST  +
        #=GF TQ  -1
        #=GF ST  127
        #=GF SP  601
        #=GF CST 135
        #=GF CSP 628
        #=GF FL  128
        #=GF MX  20p41g.matrix
        #=GF GI -25
        #=GF GE -5
    """
    if add_skip_state:
        yield get_skip_state()

    meta = {}
    seqs = []

    current_line_number = 1

    for line in file:
        if _parse_preamble_line(line):
            continue

        parts = _parse_meta_line(line)
        if parts is not None:
            meta[parts[0]] = parts[1]
            continue

        parts = _parse_alignment_line(line)
        if parts is not None:
            seqs.append(parts[1])
            continue

        if _parse_terminator_line(line):
            if len(seqs) != 2:
                raise FileFormatException(
                    file.name,
                    current_line_number,
                    "wrong number of sequences",
                )

            try:
                chrom_meta = _parse_chrom_meta(meta["TR"])
                if chrom_meta is None:
                    raise FileFormatException(
                        file.name,
                        current_line_number,
                        "missing TR field",
                    )

                chrom_name, chrom_start, chrom_stop = chrom_meta

                if meta["TQ"] == "t":
                    yield Alignment(
                        subfamily=meta["ID"],
                        chrom_name=chrom_name,
                        chrom_start=chrom_start,
                        chrom_stop=chrom_stop,
                        score=int(meta["SC"]),
                        start=int(meta["ST"]),
                        stop=int(meta["SP"]),
                        consensus_start=int(meta["CST"]),
                        consensus_stop=int(meta["CSP"]),
                        sequences=[seqs[0][::-1], seqs[1][::-1]],
                        strand=meta["SD"],
                        flank=int(meta["FL"]),
                        sub_matrix_name=meta["MX"],
                        gap_init=float(meta["GI"]),
                        gap_ext=float(meta["GE"]),
                    )
                else:
                    yield Alignment(
                        subfamily=meta["ID"],
                        chrom_name=chrom_name,
                        chrom_start=chrom_start,
                        chrom_stop=chrom_stop,
                        score=int(meta["SC"]),
                        start=int(meta["ST"]),
                        stop=int(meta["SP"]),
                        consensus_start=int(meta["CST"]),
                        consensus_stop=int(meta["CSP"]),
                        sequences=[seqs[0], seqs[1]],
                        strand=meta["SD"],
                        flank=int(meta["FL"]),
                        sub_matrix_name=meta["MX"],
                        gap_init=float(meta["GI"]),
                        gap_ext=float(meta["GE"]),
                    )
            except IndexError as e:
                raise FileFormatException(
                    file.name,
                    current_line_number,
                    "invalid TR field",
                )
            except KeyError as e:
                raise FileFormatException(
                    file.name,
                    current_line_number,
                    f"missing key: {e.args[0]}",
                )
            except ValueError as e:
                raise FileFormatException(
                    file.name,
                    current_line_number,
                    e.args[0],
                )

            meta.clear()
            seqs.clear()

        current_line_number += 1


def load_alignment_tool(file: TextIO) -> str:
    """
    Return alignment tool used.
    """

    while True:
        try:
            # skip header line
            header_line = next(file)
        except StopIteration:
            return "no alignment tool"

        try:
            line = next(file)
            if line.strip().upper().startswith("# ALIGNMENT TOOL"):
                # get alignment tool
                return line.split()[-1]
            return "no alignment tool"
        except StopIteration:
            return "no alignment tool"


class Shard(NamedTuple):
    start: int
    stop: int
    alignments: List[Alignment]


def shard_overlapping_alignments(
    alignments: Iterable[Alignment],
    shard_gap: int,
    add_skip_state: bool = True,
) -> Iterable[Shard]:
    """
    Shard the given alignments into overlapping groups, separated by no more
    than `shard_gap` nucleotides. This allows for more efficient processing
    for alignments over large regions of sequence (such as an entire genome)
    where many regions will be empty.
    Precondition: alignments are sorted by their start position and all
    alignments have start position <= stop position.

    >>> skip = get_skip_state()
    >>> a0 = Alignment("", "a", 1, 100, 0, 1, 10, 0, 0, [], "", 0, "", 0, 0)
    >>> a1 = Alignment("", "a", 1, 100, 0, 21, 30, 0, 0, [], "", 0, "", 0, 0)
    >>> shards = list(shard_overlapping_alignments([a0, a1], 10))
    >>> len(shards)
    2
    >>> shards[0].start
    1
    >>> shards[0].stop
    15
    >>> len(shards[0].alignments)
    2
    >>> shards[1].start
    16
    >>> shards[1].stop
    100
    >>> len(shards[1].alignments)
    2
    """
    shard_alignments: List[Alignment] = (
        [get_skip_state()] if add_skip_state else []
    )

    shard_start = 1
    shard_stop = None

    is_infinite_gap = shard_gap == INFINITE_SHARD_GAP

    last_alignment = None

    for alignment in alignments:
        last_alignment = alignment

        # If this is the first alignment we need to initialize
        # the stop position
        if shard_stop is None:
            shard_stop = alignment.stop

        if is_infinite_gap or alignment.start <= (shard_stop + shard_gap):
            # This alignment might extend past the previous
            # alignments in the shard, capture that here
            if shard_stop < alignment.stop:
                shard_stop = alignment.stop

            shard_alignments.append(alignment)
        else:
            # The gap between shards might be bigger than the
            # maximum shard gap, so we have to consume half of
            # the actual gap for this shard
            actual_gap = alignment.start - shard_stop
            shard_stop += int(actual_gap / 2)

            yield Shard(
                start=shard_start,
                stop=shard_stop,
                alignments=shard_alignments,
            )

            shard_start = shard_stop + 1
            shard_stop = alignment.stop

            # Note: important to create a new list here or we will
            # mutate the one we just handed back to the caller.
            shard_alignments = (
                [get_skip_state(), alignment] if add_skip_state else [alignment]
            )

    if last_alignment is None:
        raise ValueError("no alignments found")

    yield Shard(
        start=shard_start,
        stop=last_alignment.chrom_stop - last_alignment.chrom_start + 1,
        alignments=shard_alignments,
    )



from typing import Dict, List, Tuple

from .performance import timeit
from .ultra_provider import TandemRepeat


@timeit()
def calculate_repeat_scores(
    tandem_repeats: List[TandemRepeat],
    chunk_size: int,
    start_all: int,
    row_num: int,
    active_cells: Dict[int, List[int]],
    align_matrix: Dict[Tuple[int, int], float],
    consensus_matrix: Dict[Tuple[int, int], int],
    shard_start: int,
    shard_stop: int,
) -> Dict[int, float]:
    """
    Calculates score (according to ULTRA scoring) for every segment of size
    chunksize for all tandem repeats found in the target sequence.
    Scores are of the surrounding chunksize nucleotides in the sequence.

    At same time, updates AlignMatrix, ActiveCells, and ConsensusMatrix
    to include tandem repeat scores.

    input:
    tandem_repeats: list of tandem repeats from ULTRA output
    chunk_size: size of nucleotide chunks that are scored
    start_all: minimum starting nucleotide position
    row_num: number of rows in matrices
    columns: all columns that are not empty from alignment scores
    active_cells: dictionary that maps column number to a list of active rows
    align_matrix
    consensus_matrix: dictionary that maps a tuple (row, col) to consensus
    position in the subfam/query sequence
    shard_start: start seq pos of current shard
    shard_stop: stop seq pos of current shard

    output:
    repeat_scores: Hash implementation of sparse 1d array. Key is int that maps col
    index in the target sequence to it's tandem repeat score. Used in FillNodeConfidence.

    tr_info - list of dictionaries - [{'Start': num, 'Length': num, 'PositionScoreDelta': '-0.014500:1.27896'}, ...]
    >>> repeats = [TandemRepeat.from_json(m) for m in [{'Start': 5, 'Length': 4, 'PositionScoreDelta': '-0.5:1:1.5:1', 'Consensus': 'AT'}, {'Start': 10, 'Length': 8, 'PositionScoreDelta': '0:0.5:0.5:1.5:1.5:1:0.5:-0.5', 'Consensus': 'GTT'}]]
    >>> align_mat = {}
    >>> active_cols = {}
    >>> rep_scores = calculate_repeat_scores(repeats, 5, 4, 1, active_cols, align_mat, {}, 1, 30)
    >>> rep_scores
    {2: -0.5, 3: 1.0, 4: 1.5, 5: 1.0, 7: 0.0, 8: 0.5, 9: 0.5, 10: 1.5, 11: 1.5, 12: 1.0, 13: 0.5, 14: -0.5}
    >>> align_mat
    {(1, 2): 3.3333333333333335, (1, 3): 3.75, (1, 4): 3.75, (1, 5): 5.833333333333333, (2, 7): 1.6666666666666667, (2, 8): 3.125, (2, 9): 4.0, (2, 10): 5.0, (2, 11): 5.0, (2, 12): 4.0, (2, 13): 3.125, (2, 14): 1.6666666666666667}
    >>> active_cols
    {2: [0, 1], 3: [0, 1], 4: [0, 1], 5: [0, 1], 7: [0, 2], 8: [0, 2], 9: [0, 2], 10: [0, 2], 11: [0, 2], 12: [0, 2], 13: [0, 2], 14: [0, 2]}

    Shard cuts off TR from left side
    >>> repeats = [TandemRepeat.from_json(m) for m in [{'Start': 5, 'Length': 4, 'PositionScoreDelta': '-0.5:1:1.5:1', 'Consensus': 'AT'}, {'Start': 10, 'Length': 8, 'PositionScoreDelta': '0:0.5:0.5:1.5:1.5:1:0.5:-0.5', 'Consensus': 'GTT'}]]
    >>> align_mat = {}
    >>> active_cols = {}
    >>> rep_scores = calculate_repeat_scores(repeats, 5, 4, 1, active_cols, align_mat, {}, 6, 30)
    >>> rep_scores
    {3: 1.0, 4: 1.5, 5: 1.0, 7: 0.0, 8: 0.5, 9: 0.5, 10: 1.5, 11: 1.5, 12: 1.0, 13: 0.5, 14: -0.5}
    >>> align_mat
    {(1, 3): 3.75, (1, 4): 3.75, (1, 5): 5.833333333333333, (2, 7): 1.6666666666666667, (2, 8): 3.125, (2, 9): 4.0, (2, 10): 5.0, (2, 11): 5.0, (2, 12): 4.0, (2, 13): 3.125, (2, 14): 1.6666666666666667}
    >>> active_cols
    {3: [0, 1], 4: [0, 1], 5: [0, 1], 7: [0, 2], 8: [0, 2], 9: [0, 2], 10: [0, 2], 11: [0, 2], 12: [0, 2], 13: [0, 2], 14: [0, 2]}

    Shard cuts off TR from right side in first half of chunk
    >>> repeats = [TandemRepeat.from_json(m) for m in [{'Start': 5, 'Length': 4, 'PositionScoreDelta': '-0.5:1:1.5:1', 'Consensus': 'AT'}, {'Start': 10, 'Length': 8, 'PositionScoreDelta': '0:0.5:0.5:1.5:1.5:1:0.5:-0.5', 'Consensus': 'GTT'}]]
    >>> align_mat = {}
    >>> active_cols = {}
    >>> rep_scores = calculate_repeat_scores(repeats, 5, 4, 1, active_cols, align_mat, {}, 1, 11)
    >>> rep_scores
    {2: -0.5, 3: 1.0, 4: 1.5, 5: 1.0, 7: 0.0, 8: 0.5}
    >>> align_mat
    {(1, 2): 3.3333333333333335, (1, 3): 3.75, (1, 4): 3.75, (1, 5): 5.833333333333333, (2, 7): 1.6666666666666667, (2, 8): 3.125}
    >>> active_cols
    {2: [0, 1], 3: [0, 1], 4: [0, 1], 5: [0, 1], 7: [0, 2], 8: [0, 2]}
    """

    repeat_scores: Dict[
        int, float
    ] = {}  # maps col in target seq to ultra output score
    # for each tandem repeat, get overlapping windowed score
    for i in range(len(tandem_repeats)):
        # get repeat info
        rep = tandem_repeats[i]
        col_index = rep.start - start_all + 1  # can't be less than 1
        length = rep.length
        pos_scores = rep.position_scores

        # calc score for first chunk
        score: float = 0
        j = 0
        k = int((chunk_size - 1) / 2)
        # update repeat_scores for positions 0 to 15 if in boundary
        while j <= k and j < length:
            score += pos_scores[j]
            if shard_start <= rep.start + j <= shard_stop:
                repeat_scores[col_index + j] = pos_scores[j]
            j += 1
        window_size = j

        # check TR start >= chunk start
        # update AlignMatrix and ConsensusMatrix
        if rep.start >= shard_start:
            align_matrix[i + row_num, col_index] = (
                score * chunk_size / window_size
            )
            consensus_matrix[i + row_num, col_index] = rep.start
            # update active_cells
            if col_index in active_cells:
                active_cells[col_index].append(i + row_num)
            else:
                active_cells[col_index] = [0, i + row_num]

        # calc score for the rest of the chunks
        tr_chunk_length = length
        if rep.stop >= shard_stop:
            # TR ends outside of chunk boundary
            tr_chunk_length = shard_stop - rep.start + 1
        for j in range(1, tr_chunk_length):
            # check to remove last score in window
            if j - k - 1 >= 0:
                score -= float(pos_scores[j - k - 1])
                window_size -= 1
            # check to add new score in window
            if j + k < len(pos_scores):
                score += float(pos_scores[j + k])
                window_size += 1
                if shard_start <= rep.start + j + k <= shard_stop:
                    # add new score to repeat_scores
                    repeat_scores[col_index + j + k] = float(pos_scores[j + k])
            # Update matrices if in chunk boundary
            if rep.start + j >= shard_start:
                align_matrix[i + row_num, col_index + j] = (
                    score * chunk_size / window_size
                )
                consensus_matrix[i + row_num, col_index + j] = rep.start + j
                # add active cells
                if col_index + j in active_cells:
                    active_cells[col_index + j].append(i + row_num)
                else:
                    active_cells[col_index + j] = [0, i + row_num]
    return repeat_scores



from typing import List

from .performance import timeit


@timeit()
def extract_nodes(
    nodes: int,
    columns: List[int],
    changes_position: List[int],
    path_graph: List[int],
) -> None:
    """
    finds nodes that only have one (or less) incoming and one (or less) outgoing edge and adds
    them to remove_starts and remove_stops so they can be extracted from the alignment - all columns
    in removed nodes are removed from NonEmptyColumns (columns) so during next round of DP calculations,
    those nodes are ignored and surrounding nodes can be stitched if necessary,

    input:
    num_col: number of columns in matrices
    nodes: number of nodes in graph
    columns: non empty columns in matrices
    changes_position: node boundaries
    path_graph: 2D array that represents edges in the graph

    output:
    updated_num_col: updated number of columns after nodes are extracted

    >>> non_cols = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> change_pos = [0, 3, 7]
    >>> path_graph = [0, 1, 1, 0, 0, 1, 0, 0, 0]
    >>> extract_nodes(3, non_cols, change_pos, path_graph)
    >>> non_cols
    [0, 1, 2, 7, 8, 9]
    """

    remove_starts: List[int] = []
    remove_stops: List[int] = []

    # boolean for which nodes will be removed
    remove_nodes: List[bool] = [False for _ in range(nodes)]

    # extracting nodes that only have one incoming and one outgoing edge
    num_edges_in: List[int] = [0 for _ in range(nodes)]
    num_edges_out: List[int] = [0 for _ in range(nodes)]

    for row in range(nodes):
        for col in range(nodes):
            num_edges_in[col] += path_graph[row * nodes + col]
            num_edges_out[row] += path_graph[row * nodes + col]

    for node in range(nodes - 1):
        if num_edges_in[node] <= 1 and num_edges_out[node] <= 1:
            remove_starts.append(changes_position[node])
            remove_stops.append(changes_position[node + 1])
            remove_nodes[node] = True

    # deals with last node, so when NumNodes-1 the last remove stop is the end of the matrix
    if num_edges_in[nodes - 1] <= 1 and num_edges_out[nodes - 1] <= 1:
        remove_starts.append(changes_position[nodes - 1])
        remove_stops.append(len(columns) - 1)
        remove_nodes[nodes - 1] = True

    # removing inserted elements from NonEmptyColumns so they can be ignored
    total: int = 0
    for i in range(len(remove_stops)):
        del columns[remove_starts[i] - total : remove_stops[i] - total]
        # 	helps with offset, when first part is spliced out need an offset to know where to splice out for second part
        total += remove_stops[i] - remove_starts[i]



from typing import Dict, List

from .alignment import Alignment

ConsensusLengths = Dict[str, int]


def find_consensus_lengths(alignments: List[Alignment]) -> ConsensusLengths:
    """
    Consensus sequence lengths used by the SODA printer.
    """
    consensus_lengths = {}
    for a in alignments:
        flank = a.flank

        if a.strand == "+":
            stop_position = a.consensus_stop
            consensus_lengths[a.subfamily] = stop_position + flank
        else:
            start_position = a.consensus_start
            consensus_lengths[a.subfamily] = start_position + flank

    return consensus_lengths



import sys
import re


def read_file(filename_cm):
    """
    opens file and returns contents of file as one string
    """
    file_contents = ""
    with open(filename_cm, "r") as f_cm:
        file_contents = f_cm.read()

    return file_contents


def get_info(info_array):
    """
    takes info line split by white space and assigns all info fields to variables
    """
    score = info_array[0]
    chrom = info_array[4]
    chrom = info_array[4]

    # doesn't have chrom info because it's an artificial seq
    if "chr" not in chrom:
        chrom = "chr0:0000-0000"

    start = info_array[5]
    stop = info_array[6]
    subfam = ""
    consensus_start = ""
    consensus_stop = ""
    strand = ""
    flank = ""

    # if 6 contains (), then start/stop = 7/8
    if "(" in info_array[6]:  # FIXME - shouldn't this be 5?
        start = info_array[6]
        stop = info_array[7]

    if len(info_array) == 12:  # not compliment
        strand = "+"
        subfam = info_array[8]
        consensus_start = info_array[9]
        consensus_stop = info_array[10]
        flank = info_array[11]

        # if 10 contains (), then start/stop = 7/8
        if "(" in info_array[9]:
            flank = info_array[9]
            start = info_array[10]
            stop = info_array[11]

    else:  # compliment
        strand = "-"
        subfam = info_array[9]
        consensus_start = info_array[10]
        consensus_stop = info_array[11]
        flank = info_array[12]

        if "(" in info_array[10]:
            flank = info_array[10]
            consensus_start = info_array[11]
            consensus_stop = info_array[12]

    flank = flank.strip(")")
    flank = flank.strip("(")

    return (
        subfam,
        chrom,
        score,
        strand,
        start,
        stop,
        consensus_start,
        consensus_stop,
        flank,
    )


def get_score_matrix(file_contents):
    """
    grabs score matrix info from alignment file, puts it in correct format, returns string
    """
    score_matrix = ""
    m_string = re.search(r"Score matrix.+?\n([\s\S]+?)\n\n", file_contents)
    m_array = m_string[1].split("\n")

    score_matrix += m_array[0].lstrip() + "\n"
    for i in range(1, len(m_array)):
        m_array[i] = m_array[i][1::]
        score_matrix += m_array[i].lstrip() + "\n"

    return score_matrix


def get_gap_penalties(file_contents):
    """
    returns gap_init and gap_ext
    """
    m_string = re.search(
        r"Gap penalties: gap_init: (.+?), gap_ext: (.+?),", file_contents
    )
    gap_init = m_string[1]
    gap_ext = m_string[2]

    return gap_init, gap_ext


def print_info(
    C,
    subfam,
    chrom,
    score,
    strand,
    start,
    stop,
    consensus_start,
    consensus_stop,
    flank,
    matrix_name,
    gap_init,
    gap_ext,
    f_out_sto,
):
    """
    prints all info in correct format for stockholm
    """
    f_out_sto.write(f"#=GF ID  {subfam}\n")
    f_out_sto.write(f"#=GF TR  {chrom}\n")
    f_out_sto.write(f"#=GF SC  {score}\n")
    f_out_sto.write(f"#=GF SD  {strand}\n")

    if strand == "-":
        if C:
            f_out_sto.write(f"#=GF TQ  t\n")
        else:
            f_out_sto.write(f"#=GF TQ  q\n")
    else:
        f_out_sto.write(f"#=GF TQ  -1\n")

    f_out_sto.write(f"#=GF ST  {start}\n")
    f_out_sto.write(f"#=GF SP  {stop}\n")
    f_out_sto.write(f"#=GF CST {consensus_start}\n")
    f_out_sto.write(f"#=GF CSP {consensus_stop}\n")
    f_out_sto.write(f"#=GF FL  {flank}\n")
    f_out_sto.write(f"#=GF MX  {matrix_name}\n")
    f_out_sto.write(f"#=GF GI  {gap_init}\n")
    f_out_sto.write(f"#=GF GE  {gap_ext}\n")


def get_alignment(alignment_array):
    """
    takes cm alignment split on newline and return string of chrom seq and subfam seq
    """
    chrom_seq = ""
    subfam_seq = ""

    # for (my $i = 0; $i < @ Alignment; $i = $i + 4){
    for i in range(0, len(alignment_array), 4):
        m_chrom = re.search(r".+?\s\d+\s(.+?)\s\d+", alignment_array[i])
        if m_chrom:
            chrom_seq += m_chrom.group(1)

        m_subfam = re.search(r".+?\s\d+\s(.+?)\s\d+", alignment_array[i + 2])
        if m_chrom:
            subfam_seq += m_subfam.group(1)

    return chrom_seq, subfam_seq


def print_alignment(chrom_seq, subfam_seq, chrom, subfam, f_out_sto):
    """
    takes chrom and subfam names and chrom and subfam seqs and prints in
    stockholm format
    """
    f_out_sto.write(f"{chrom}   {chrom_seq}\n")
    f_out_sto.write(f"{subfam}   {subfam_seq}\n")

    f_out_sto.write("//\n")


def print_score_matrix(
    filename_out_matrix, score_matrix, matrix_name, background_freqs
):
    """
    print score matrix to its own output file with extension ".matrix"
    """
    f_out_matrix = open(filename_out_matrix, "w")
    f_out_matrix.write(matrix_name)
    f_out_matrix.write("\n")
    f_out_matrix.write(f"BACKGROUND FREQS: {background_freqs}")
    f_out_matrix.write("\n")
    f_out_matrix.write(score_matrix)
    f_out_matrix.write("//\n")
    f_out_matrix.close()


def main():
    filename_cm = sys.argv[1]
    file_contents = read_file(filename_cm)

    filename_out_sto = filename_cm + ".sto"
    f_out_sto = open(filename_out_sto, "w")

    filename_out_matrix = filename_cm + ".matrix"
    matrix_name = "matrix1"

    f_out_sto.write("# STOCKHOLM 1.0\n")
    f_out_sto.write(f"# ALIGNMENT TOOL cross_match\n")

    # get background freqs
    background_freqs = re.findall(
        r"Assumed background frequencies:.*\n.*\n.*",
        file_contents,
    )
    background_freqs = background_freqs[0].splitlines()[1].split()
    background_freqs = background_freqs[: len(background_freqs) - 2]
    background_freqs_dict = {}
    for i in range(0, len(background_freqs), 2):
        char = background_freqs[i][0]
        freq = float(background_freqs[i + 1])
        if freq != 0:
            background_freqs_dict[char] = freq

    score_matrix = get_score_matrix(file_contents)
    print_score_matrix(
        filename_out_matrix, score_matrix, matrix_name, background_freqs_dict
    )

    gap_init, gap_ext = get_gap_penalties(file_contents)

    alignments = re.findall(
        r"\s*?\d+\s+[0-9]+\.[0-9]+\s+[0-9.]+\s+[0-9.]+\s+.+?\n\n[\s\S]+?Transitions",
        file_contents,
    )

    for region in alignments:
        info_line: str = ""
        alignment: str = ""
        region = region.strip()
        m = re.match(r"(.+?)\n([\s\S]+)\n\nTransitions", region)
        if m:
            info_line = m.group(1)
            alignment = m.group(2)

        alignment = alignment.strip("\n")

        info_array = info_line.split()

        (
            subfam,
            chrom,
            score,
            strand,
            start,
            stop,
            consensus_start,
            consensus_stop,
            flank,
        ) = get_info(info_array)

        alignment_array = alignment.split("\n")

        # if C is true: on reverse strand and target is reversed
        C = 0
        if "C" in alignment_array[0]:
            C = 1

        chrom_seq, subfam_seq = get_alignment(alignment_array)

        print_info(
            C,
            subfam,
            chrom,
            score,
            strand,
            start,
            stop,
            consensus_start,
            consensus_stop,
            flank,
            matrix_name,
            gap_init,
            gap_ext,
            f_out_sto,
        )
        print_alignment(chrom_seq, subfam_seq, chrom, subfam, f_out_sto)

    f_out_sto.close()


if __name__ == "__main__":
    main()



from typing import Dict, List, Tuple

from .matrices import ConfidenceMatrix, SupportMatrix
from .performance import timeit


@timeit()
def fill_support_matrix(
    chunk_size: int,
    starts: List[int],
    stops: List[int],
    confidence_matrix: ConfidenceMatrix,
) -> SupportMatrix:
    """
    Fill the support_matrix using values in `confidence_matrix`. Average
    confidence values for surrounding `chunk_size` confidence values -
    normalized by dividing by number of segments.

    The score for subfam x at position i is the sum of all confidences for
    subfam x for all segments that overlap position i - divided by the number of
    segments.

    Inputs:

    chunk_size - the width of the window to use when averaging confidence
    scores. This is assumed to be an odd number >= 3.

    starts - a list of start columns for each row in the confidence matrix.
    These positions come should come directly from the alignments without any
    offsetting. The exception is the first value, which should contain the
    minimum of the remaining values, less one, in other words, `starts[0] ==
    min(starts[1:]) - 1`. This is the implicit start position for the skip
    state, which occupies the first row and is the only row that has data in the
    first column of the confidence matrix.

    stops - equivalent to the starts list, but contains the last column in each
    row that contains data. Again, the first value is special and should contain
    the maximum of the remaining values. In other words, `stops[0] ==
    max(stops[1:]) + 1`.

    confidence_matrix - confidence values computed with fill_confidence_matrix.
    The confidence matrix should have one contiguous run of values per row.

    Outputs:

    support_matrix - support values represent average confidence values over a
    window chunk_size wide at each position in the confidence matrix. This is
    the "support score".

    >>> conf_mat = {
    ...     (0, 0): 1.0, (0, 1): 0.1, (0, 2): 0.2, (0, 3): 0.4, (0, 4): 0.5, (0, 5): 1.0,
    ...     (1, 1): 0.9, (1, 2): 0.5, (1, 3): 0.5, (1, 4): 0.5,
    ...     (2, 2): 0.3, (2, 3): 0.1,
    ... }
    >>> supp_mat = fill_support_matrix(
    ...     chunk_size=3,
    ...     starts=[0, 1, 2],
    ...     stops=[5, 4, 3],
    ...     confidence_matrix=conf_mat)
    >>> len(supp_mat) == len(conf_mat)
    True
    >>> round(supp_mat[1, 1], 4)
    0.7
    >>> round(supp_mat[1, 2], 4)
    0.6333
    >>> round(supp_mat[1, 3], 4)
    0.5
    >>> round(supp_mat[1, 4], 4)
    0.5
    >>> (1, 0) in supp_mat
    False
    >>> round(supp_mat[2, 2], 4)
    0.2
    >>> round(supp_mat[2, 3], 4)
    0.2
    >>> (2, 4) in supp_mat
    False
    >>>
    >>> conf_mat = {
    ...     (0, 0): 1.0, (0, 1): 0.1, (0, 2): 0.05, (0, 3): 0.01, (0, 4): 0.01, (0, 5): 0.01, (0, 6): 0.9, (0, 7): 1.0,
    ...     (1, 1): 0.9, (1, 2): 0.75, (1, 3): 0.5, (1, 4): 0.39, (1, 5): 0.1, (1, 6): 0.1,
    ...     (2, 2): 0.2, (2, 3): 0.49, (2, 4): 0.6, (2, 5): 0.89,
    ... }
    >>> supp_mat = fill_support_matrix(
    ...     chunk_size=3,
    ...     starts=[9, 10, 11],
    ...     stops=[16, 15, 14],
    ...     confidence_matrix=conf_mat)
    >>> len(supp_mat) == len(conf_mat)
    True
    >>> supp_mat[1, 1]
    0.825
    >>> round(supp_mat[1, 2], 4)
    0.7167
    >>> round(supp_mat[1, 3], 4)
    0.5467
    >>> round(supp_mat[1, 4], 4)
    0.33
    >>> round(supp_mat[1, 5], 4)
    0.1967
    >>> round(supp_mat[1, 6], 4)
    0.1
    >>> (2, 1) in supp_mat
    False
    >>> round(supp_mat[2, 2], 4)
    0.345
    >>> round(supp_mat[2, 3], 4)
    0.43
    >>> round(supp_mat[2, 4], 4)
    0.66
    >>> round(supp_mat[2, 5], 4)
    0.745
    >>> (2, 6) in supp_mat
    False
    """

    support_matrix: Dict[Tuple[int, int], float] = {}

    half_chunk: int = int((chunk_size - 1) / 2)
    row_count = len(starts)

    # This is the amount by which we need to slide all the start and stop values
    # over to the left to match the column values in the confidence matrix for each
    # subfamily alignment.
    # See the documentation for `start` on the `Alignment` class.
    # FIXME: In some cases this may not be set correctly - how so??
    # This is a problem, the tests don't cover this and they only pass because
    # the value is zero in both cases.
    position_offset = starts[0]  # start of skip state

    # Note: start and stop values are closed ranges, so when a stop value is
    # used in a range() call, the stop value passed to range needs to be
    # incremented by 1. For example, range(start, stop + 1). This convention is
    # consistent through the algorithm below.

    for row_index in range(0, row_count):
        start: int = starts[row_index] - position_offset
        stop: int = stops[row_index] - position_offset

        prev_chunk_start = start
        prev_chunk_stop = stop

        sum_of_scores = 0.0

        for col_index in range(start, stop + 1):
            chunk_start = max(col_index - half_chunk, start)
            chunk_stop = min(col_index + half_chunk, stop)

            column_count = chunk_stop - chunk_start + 1

            if col_index == start:
                # Compute the entire chunk if we're working on the first column
                # in the row. This bootstraps our more efficient summation
                # later.
                sum_of_scores = 0.0
                for sum_index in range(chunk_start, chunk_stop + 1):
                    sum_of_scores += confidence_matrix[row_index, sum_index]
            else:
                # Subtract the value from the left-most column, but only if it
                # has slipped out of our window. Add the value from the
                # right-most column, but only if it is newly part of our window.
                if chunk_start > prev_chunk_start:
                    sum_of_scores -= confidence_matrix[
                        row_index, prev_chunk_start
                    ]
                if chunk_stop > prev_chunk_stop:
                    sum_of_scores += confidence_matrix[row_index, chunk_stop]

            support_matrix[row_index, col_index] = sum_of_scores / column_count

            prev_chunk_start = chunk_start
            prev_chunk_stop = chunk_stop

    return support_matrix



import logging
from sys import argv, stderr
from typing import List

from ._options import Options
from ._runners import run_confidence, run_full
from .lambda_provider import EaselLambdaProvider
from .load_alignments import (
    load_alignments,
    load_alignment_tool,
    shard_overlapping_alignments,
)
from .output import Output
from .printers import Printer
from .prior_counts import read_prior_counts
from .substitution_matrix import load_substitution_matrices
from .ultra_provider import ApplicationUltraProvider, TandemRepeat


class AppError(RuntimeError):
    pass


def _configure_logging(opts: Options) -> None:
    if opts.log_file_path != "":
        log_file = open(opts.log_file_path, "w")
    else:
        log_file = stderr
    handler = logging.StreamHandler(log_file)

    formatter = logging.Formatter(
        fmt="{levelname} ({name}) - {message}", style="{"
    )
    handler.setFormatter(formatter)

    logger = logging.getLogger(__package__)
    logger.addHandler(handler)

    if opts.log_level == "debug":
        logger.setLevel(logging.DEBUG)
    elif opts.log_level == "verbose":
        logger.setLevel(logging.INFO)
    elif opts.log_level == "normal":
        logger.setLevel(logging.WARNING)
    elif opts.log_level == "quiet":
        logger.setLevel(logging.CRITICAL)


def _configure_tandem_repeats(opts: Options) -> List[TandemRepeat]:
    if opts.ultra_data_path and opts.sequence_file_path:
        # This is ambiguous so we bail out
        raise AppError("cannot specify both ultra data and sequence files")

    tandem_repeats = []

    if opts.ultra_data_path or opts.sequence_file_path:
        # Run ULTRA or load TRs from an output file
        provider = ApplicationUltraProvider(
            opts.sequence_file_path,
            opts.ultra_data_path,
            opts.ultra_path,
        )
        ultra_output = provider()
        tandem_repeats = ultra_output.tandem_repeats

    return tandem_repeats


def run():
    opts = Options(argv[1:])
    _configure_logging(opts)

    # ----------------------------
    # Tandem Repeat initialization
    # ----------------------------

    # This will come back as an empty list if we don't need to worry about TRs.
    tandem_repeats = _configure_tandem_repeats(opts)

    # -----------------
    # Sub-family counts
    # -----------------

    infile_prior_counts = (
        open(opts.prior_counts_path, "r") if opts.prior_counts_path else None
    )

    subfam_counts = (
        read_prior_counts(infile_prior_counts, bool(tandem_repeats))
        if infile_prior_counts is not None
        else {}
    )

    # ----------------------------
    # Load the substitution matrix
    # ----------------------------

    with open(opts.alignments_file_path) as _align_tool_infile:
        alignment_tool: str = load_alignment_tool(_align_tool_infile)

    # Note: alignments from blast or HMMER are not set-up to use complexity
    # adjusted scoring
    if opts.complexity_adjustment and alignment_tool not in [
        "cross_match",
        "RepeatMasker",
    ]:
        raise AppError(
            f"cannot use complexity adjusted scoring with {alignment_tool}"
        )

    _lambda_provider = EaselLambdaProvider(opts.easel_path)
    with open(opts.sub_matrices_path) as _sub_matrices_file:
        sub_matrices = load_substitution_matrices(
            _sub_matrices_file, _lambda_provider, opts.complexity_adjustment
        )

    # -------------------------------------------------
    # Flags and parameters related to secondary outputs
    # -------------------------------------------------

    outputter = Output(opts.output_path, opts.output_to_file)

    # -----------------------------
    # Load alignments to operate on
    # -----------------------------
    with open(opts.alignments_file_path) as _infile:
        alignments = list(load_alignments(_infile))

    # --------------------------
    # Run confidence calculation
    # --------------------------

    lambda_values = [sub_matrices[a.sub_matrix_name].lamb for a in alignments]

    if opts.confidence:
        run_confidence(
            alignments,
            lambs=lambda_values,
        )
        exit()

    # ---------------------
    # Set up general output
    # ---------------------

    results_file = outputter.get_results()
    printer = Printer(
        output_file=results_file,
        print_id=opts.ids,
        use_matrix_position=opts.matrix_position,
        use_sequence_position=opts.sequence_position,
    )
    printer.print_results_header()

    # ----------------------------------------------------------------
    # Loop through the alignment shards and process each independently
    # ----------------------------------------------------------------

    # FIXME: if shard gap is infinite, use all TRs (skip breaking them up)
    tr_start: int = 0
    _prev_start: int = -1
    _prev_stop: int = -1

    for index, chunk in enumerate(
        shard_overlapping_alignments(alignments, shard_gap=opts.shard_gap)
    ):
        chunk_start = chunk.start
        chunk_stop = chunk.stop
        # get TRs between chunk stop and start
        tr_end = tr_start
        while tr_end < len(tandem_repeats):
            tr = tandem_repeats[tr_end]
            if tr.start <= chunk_stop:
                tr_end += 1
            else:
                break
        tandem_repeats_chunk = tandem_repeats[tr_start:tr_end]
        tr_start = tr_end
        if tr_start > 0 and tandem_repeats[tr_start - 1].stop > chunk_stop:
            tr_start -= 1

        soda_viz_file, soda_conf_file = (
            outputter.get_soda(index) if opts.soda else (None, None)
        )
        printer.set_soda_files(soda_viz_file, soda_conf_file)

        (_last_start, _last_stop) = run_full(
            chunk.alignments,
            tandem_repeats_chunk,
            opts.chunk_size,
            sub_matrices,
            subfam_counts,
            chunk_start,
            chunk_stop,
            _prev_start,
            _prev_stop,
            printer,
        )
        _prev_start, _prev_stop = (
            _last_start,
            _last_stop,
        )

        if soda_viz_file is not None:
            soda_viz_file.close()
        if soda_conf_file is not None:
            soda_conf_file.close()



import json
import os
from logging import Logger
from typing import Any, Callable, Dict, List, NamedTuple, Tuple

from .performance import timeit


class TandemRepeat(NamedTuple):
    consensus: str
    start: int
    length: int
    stop: int
    position_scores: Tuple[float, ...]

    @staticmethod
    def from_json(json_map: Dict[str, Any]):
        raw_scores = json_map["PositionScoreDelta"].split(":")
        position_scores = []
        for score in raw_scores:
            if abs(float(score)) > 100:
                position_scores.append(0.0)
                Logger(__name__).warning(
                    """
                    Unreasonably-large score detected from ULTRA, 
                    in sequence region {}..{}, score={}
                    """.format(
                        int(json_map["Start"]),
                        int(json_map["Start"]) + int(json_map["Length"]) - 1,
                        score,
                    )
                )
            else:
                position_scores.append(float(score))

        return TandemRepeat(
            consensus=json_map["Consensus"],
            start=int(json_map["Start"]),
            length=int(json_map["Length"]),
            stop=int(json_map["Start"]) + int(json_map["Length"]) - 1,
            position_scores=tuple(position_scores),
        )


class UltraOutput(NamedTuple):
    tandem_repeats: List[TandemRepeat]

    @staticmethod
    def from_json(json_map: Dict[str, Any]):
        repeats = list(map(TandemRepeat.from_json, json_map["Repeats"]))
        return UltraOutput(tandem_repeats=repeats)

    @property
    def tr_count(self) -> int:
        return len(self.tandem_repeats)


UltraProvider = Callable[[], UltraOutput]


class ApplicationUltraProvider:
    _sequence_path: str
    _ultra_output_path: str
    _ultra_path: str

    def __init__(
        self,
        sequence_path: str = "",
        ultra_output_path: str = "",
        ultra_path: str = "ultra",
    ):
        if ultra_output_path == "" and sequence_path == "":
            raise RuntimeError(
                "must provide either ultra output path or sequence file path"
            )

        self._sequence_path = sequence_path
        self._ultra_output_path = ultra_output_path
        self._ultra_path = ultra_path

    @timeit()
    def __call__(self) -> UltraOutput:
        if self._sequence_path:
            ultra_stream = os.popen(
                self._ultra_path + " -ss " + self._sequence_path
            )
            raw_output = json.load(ultra_stream)
        else:
            with open(self._ultra_output_path, "r") as output_file:
                raw_output = json.load(output_file)

        ultra_output = UltraOutput.from_json(raw_output)

        return ultra_output



from logging import Logger
from math import log
from sys import stdout
from typing import Dict, List, Optional, TextIO, Tuple

from .alignment import Alignment
from .calc_repeat_scores import calculate_repeat_scores
from .collapse_matrices import collapse_matrices
from .confidence_cm import confidence_only
from .constants import CHANGE_PROB, SAME_PROB_LOG, SKIP_ALIGN_SCORE
from .edges import edges
from .extract_nodes import extract_nodes
from .fill_align_matrix import fill_align_matrix
from .fill_confidence_matrix import fill_confidence_matrix
from .fill_consensus_position_matrix import fill_consensus_position_matrix
from .fill_node_confidence import fill_node_confidence
from .fill_path_graph import fill_path_graph
from .fill_probability_matrix import fill_probability_matrix
from .fill_support_matrix import fill_support_matrix
from .get_path import get_path
from .pad_sequences import pad_sequences
from .print_helpers import find_consensus_lengths
from .printers import Printer
from .substitution_matrix import SubMatrix, SubMatrixCollection
from .ultra_provider import TandemRepeat


def run_confidence(
    alignments: List[Alignment],
    lambs: List[float],
) -> None:
    # command line option to just output confidence values for
    # single annotation instead of do whole algorithm
    """
    Calculate confidence values for single annotations instead of
    running the entire algorithm.

    :param alignments: list of alignments to run on
    :param lambs: the values of lambda to use for each alignment (from Easel)
    """
    if len(alignments) != len(lambs):
        raise RuntimeError(
            "number of alignments must match number of lambda values"
        )

    subfams = []
    subfam_rows = []
    scores = []

    for i, a in enumerate(alignments):
        subfams.append(a.subfamily)
        subfam_rows.append(i)
        scores.append(a.score)

    confidence_list = confidence_only(scores, lambs)

    # ignore this because the types work, but mypy doesn't know that
    confidence_list, subfams_copy = zip(*sorted(zip(confidence_list, subfams)))  # type: ignore

    stdout.write(f"query_label\tconfidence\n")
    for i in range(len(subfams_copy) - 1, 0, -1):
        if confidence_list[i]:
            stdout.write(f"{subfams_copy[i]}\t{confidence_list[i]}\n")


def _validate_target(target: Alignment) -> None:
    if target.chrom_length == 0:
        Logger(__name__).warning(
            """No chromosome position information found
            (this is OK for artificial sequences but --viz and --heatmap will fail)"""
        )
    elif target.chrom_length <= 25:
        raise RuntimeError("Target sequence length must be >25")
    elif target.chrom_length < 1000:
        Logger(__name__).warning(
            """Target region is <1000 in length,
            did you mean to use such a short region?"""
        )


def _handle_single_alignment(
    target: Alignment,
    printer: Printer,
) -> Tuple[int, int]:
    """
    If there is only one subfam in the alignment file, no need
    to run anything because we know that subfam is the annotation.
    """
    from uuid import uuid4

    node_id = uuid4().hex
    if printer.use_matrix_position:
        printer.print_results_simple(
            1,
            target.stop - target.start + 1,
            node_id,
            target.subfamily,
        )
    elif printer.use_sequence_position:
        printer.print_results_simple(
            target.start,
            target.stop,
            node_id,
            target.subfamily,
        )
    else:
        printer.print_results_simple(
            target.start + target.chrom_start,
            target.stop + target.chrom_start,
            node_id,
            target.subfamily,
        )
    return target.start, target.stop


def _change_probs(seq_count: int) -> Tuple[float, float, float]:
    change_prob_log = log(CHANGE_PROB / (seq_count - 1))

    # jumping in and then out of the skip state counts as 1 jump
    change_prob_skip = change_prob_log / 2

    # 5% of the jump penalty, staying in skip state for 20nt "counts" as one jump
    same_prob_skip = change_prob_log / 30

    return change_prob_log, change_prob_skip, same_prob_skip


def run_full(
    alignments: List[Alignment],
    tandem_repeats: List[TandemRepeat],
    chunk_size: int,
    sub_matrix_scores: SubMatrixCollection,
    subfam_counts: Dict[str, float],
    shard_start: int,
    shard_stop: int,
    prev_start: int,
    prev_stop: int,
    printer: Printer,
) -> Tuple[int, int]:
    # chunk start and stop are positions in seq
    seq_count = len(alignments)

    target = alignments[1]
    _validate_target(target)

    if len(tandem_repeats) == 0 and seq_count == 2:
        # only one alignment other than the skip state
        return _handle_single_alignment(
            target,
            printer,
        )

    change_prob_log, change_prob_skip, same_prob_skip = _change_probs(seq_count)

    # We use a whole bunch of parallel lists to store alignment data. This may
    # change in the future, but for now it is an artifact of an earlier version
    # of the code and it has some benefits in that we can more easily iterate
    # over a particular field. Each of these lists holds data for exactly one
    # alignment (one row in the matrices we build later).
    alignment_subfamilies: List[str] = []
    alignment_chromosomes: List[str] = []
    alignment_scores: List[int] = []
    alignment_strands: List[str] = []
    alignment_start_positions: List[int] = []
    alignment_stop_positions: List[int] = []
    alignment_consensus_start_positions: List[int] = []
    alignment_consensus_stop_positions: List[int] = []
    alignment_subfamily_sequences: List[str] = []
    alignment_chromosome_sequences: List[str] = []
    alignment_flanks: List[int] = []
    alignment_substitution_matrices: List[SubMatrix] = []
    alignment_gap_inits: List[float] = []
    alignment_gap_exts: List[float] = []
    alignment_lambdas: List[float] = []

    for a in alignments:
        alignment_subfamilies.append(a.subfamily)
        alignment_chromosomes.append(a.chrom_name)
        alignment_scores.append(a.score)
        alignment_strands.append(a.strand)
        alignment_start_positions.append(a.start)
        alignment_stop_positions.append(a.stop)
        alignment_consensus_start_positions.append(a.consensus_start)
        alignment_consensus_stop_positions.append(a.consensus_stop)
        alignment_subfamily_sequences.append(a.subfamily_sequence)
        alignment_chromosome_sequences.append(a.sequence)
        alignment_flanks.append(a.flank)
        alignment_gap_inits.append(a.gap_init)
        alignment_gap_exts.append(a.gap_ext)

        # fixme (george): is this a list of dicts?
        # fixme (george): change it to a dict of dicts?
        if a.sub_matrix_name in sub_matrix_scores:
            alignment_substitution_matrices.append(
                sub_matrix_scores[a.sub_matrix_name]
            )
            alignment_lambdas.append(sub_matrix_scores[a.sub_matrix_name].lamb)
        else:
            alignment_substitution_matrices.append(SubMatrix("skip", 0.0))
            alignment_lambdas.append(0.0)

    tr_consensus_matrix: List[str] = []
    if tandem_repeats:
        for tr in tandem_repeats:
            tr_consensus_matrix.append(tr.consensus)
            # add TR starts and stops
            # check they do not exceed chunk boundary
            if tr.start < shard_start:
                alignment_start_positions.append(shard_start)
            else:
                alignment_start_positions.append(tr.start)
            if tr.stop > shard_stop:
                alignment_stop_positions.append(shard_stop)
            else:
                alignment_stop_positions.append(tr.stop)

    # The implicit start and stop positions for the skip state are the index
    # prior to the minimum start and the index after the maximum stop. In other
    # words, the skip state has N+2 columns, where N is the number of columns
    # spanned by the alignments.
    alignment_start_positions[0] = min(alignment_start_positions[1:]) - 1
    alignment_stop_positions[0] = max(alignment_stop_positions[1:]) + 1

    # save original alignments before padding for SODA viz
    original_subfamily_sequences = list(alignment_subfamily_sequences)
    original_chromosome_sequences = list(alignment_chromosome_sequences)

    pad_sequences(
        chunk_size,
        alignment_subfamily_sequences,
        alignment_chromosome_sequences,
    )

    start_all, stop_all = edges(
        alignment_start_positions, alignment_stop_positions
    )

    # Plus one because the range is inclusive, plus two because
    # we have the implicit first and last positions that only
    # have data for the skip state.
    column_count = stop_all - start_all + 1 + 2

    # number of rows in matrices
    row_count: int = seq_count

    align_matrix = fill_align_matrix(
        alignment_lambdas,
        column_count,
        start_all,
        chunk_size,
        alignment_gap_inits,
        alignment_gap_exts,
        SKIP_ALIGN_SCORE,
        alignment_subfamily_sequences,
        alignment_chromosome_sequences,
        alignment_start_positions,
        alignment_stop_positions,
        [sm.scores for sm in alignment_substitution_matrices],
        [sm.background_freqs for sm in alignment_substitution_matrices],
    )

    (active_cells, consensus_matrix) = fill_consensus_position_matrix(
        row_count,
        column_count,
        start_all,
        alignment_subfamily_sequences,
        alignment_chromosome_sequences,
        alignment_start_positions,
        alignment_stop_positions,
        alignment_consensus_start_positions,
        alignment_strands,
    )

    repeat_scores: Dict[int, float] = {}
    if len(tandem_repeats) > 0:
        repeat_scores = calculate_repeat_scores(
            tandem_repeats,
            chunk_size,
            start_all,
            row_count,
            active_cells,
            align_matrix,
            consensus_matrix,
            shard_start,
            shard_stop,
        )

        # add TRs to subfams
        for _ in tandem_repeats:
            alignment_subfamilies.append("Tandem Repeat")
            alignment_strands.append("+")
            row_count += 1

    confidence_matrix = fill_confidence_matrix(
        column_count,
        subfam_counts,
        alignment_subfamilies,
        active_cells,
        align_matrix,
    )

    support_matrix = fill_support_matrix(
        chunk_size,
        alignment_start_positions,
        alignment_stop_positions,
        confidence_matrix,
    )

    non_empty_columns = [c for c in range(column_count)]

    collapsed_matrices = collapse_matrices(
        row_count,
        start_all,
        non_empty_columns,
        alignment_subfamilies,
        alignment_strands,
        alignment_start_positions,
        alignment_stop_positions,
        active_cells,
        support_matrix,
        consensus_matrix,
    )

    support_matrix_collapse = collapsed_matrices.support_matrix
    subfams_collapse = collapsed_matrices.subfamilies
    active_cells_collapse = collapsed_matrices.active_rows
    consensus_matrix_collapse = collapsed_matrices.consensus_matrix
    strand_matrix_collapse = collapsed_matrices.strand_matrix
    subfams_collapse_index = collapsed_matrices.subfamily_indices
    row_count = collapsed_matrices.row_num_update
    subfam_alignments_collapse = collapsed_matrices.subfam_alignments_matrix

    align_matrix.clear()
    confidence_matrix.clear()
    support_matrix.clear()
    consensus_matrix.clear()

    if tandem_repeats:
        # give different TRs consensus positions that don't allow them to be stitched
        tr_consensus_pos = 1000000
        prev_tr_col = 0
        for tr_col in repeat_scores:
            if prev_tr_col != tr_col - 1:
                tr_consensus_pos -= 500
            consensus_matrix_collapse[
                row_count - 1, tr_col + 1
            ] = tr_consensus_pos
            prev_tr_col = tr_col

    (
        prob_matrix_last_column,
        origin_matrix,
        same_subfam_change_matrix,
    ) = fill_probability_matrix(
        same_prob_skip,
        SAME_PROB_LOG,
        change_prob_log,
        change_prob_skip,
        non_empty_columns,
        collapsed_matrices,
    )

    # node_ids for each nucleotide will be assigned during DP backtrace
    node_ids = [""] * column_count

    (changes_position, changes) = get_path(
        non_empty_columns,
        node_ids,
        subfams_collapse,
        prob_matrix_last_column,
        active_cells_collapse,
        origin_matrix,
        same_subfam_change_matrix,
    )

    # keep the original annotation for reporting results
    changes_orig = changes.copy()
    changes_position_orig = changes_position.copy()
    non_empty_columns_orig = non_empty_columns.copy()

    node_confidence: Dict[Tuple[str, int], float] = {}
    node_confidence_orig: Dict[Tuple[str, int], float] = {}
    path_graph: List[int] = []

    # Finding inserted elements and stitching original elements
    # Steps-
    # 1. Find confidence for nodes
    # 2. Create path graph - Find alternative paths through graph and add those edges
    # 3. Extract all nodes (from dp matrix) that have a single incoming and a single outgoing edge
    # 5. Annotate again with removed nodes
    #   ** stop when all nodes have incoming and outgoing edges <= 1 or there are <= 2 nodes left
    prev_num_nodes: int = 0
    count: int = 0
    while True:
        count += 1
        node_count = len(changes)

        node_confidence.clear()  # reuse old node_confidence matrix
        node_confidence = fill_node_confidence(
            node_count,
            start_all,
            alignment_gap_inits,
            alignment_gap_exts,
            non_empty_columns,
            alignment_start_positions,
            alignment_stop_positions,
            changes_position,
            alignment_subfamilies,
            alignment_subfamily_sequences,
            alignment_chromosome_sequences,
            subfam_counts,
            alignment_substitution_matrices,
            repeat_scores,
            len(tandem_repeats),
        )

        # store original node confidence for reporting results
        if count == 1:
            node_confidence_orig = node_confidence.copy()

        # breakout of loop if there are 2 or less nodes left
        if node_count <= 2 or node_count == prev_num_nodes:
            break

        path_graph.clear()  # reuse old path_graph

        path_graph = fill_path_graph(
            node_count,
            non_empty_columns,
            changes,
            changes_position,
            consensus_matrix_collapse,
            strand_matrix_collapse,
            node_confidence,
            subfams_collapse_index,
        )

        # test to see if there are nodes in the graph that have more than one incoming or outgoing edge,
        # if so - keep looping, if not - break out of the loop
        # if they are all 0, break out of the loop
        test: bool = False
        for j in range(node_count):
            for i in range(j - 1):
                if path_graph[i * node_count + j] == 1:
                    test = True

        if not test:
            break

        extract_nodes(
            node_count,
            non_empty_columns,
            changes_position,
            path_graph,
        )

        # run DP calculations again with nodes corresponding to inserted elements removed
        # ignores removed nodes because they are no longer in NonEmptyColumns
        (
            prob_matrix_last_column,
            origin_matrix,
            same_subfam_change_matrix,
        ) = fill_probability_matrix(
            same_prob_skip,
            SAME_PROB_LOG,
            change_prob_log,
            change_prob_skip,
            non_empty_columns,
            collapsed_matrices,
        )

        (changes_position, changes) = get_path(
            non_empty_columns,
            node_ids,
            subfams_collapse,
            prob_matrix_last_column,
            active_cells_collapse,
            origin_matrix,
            same_subfam_change_matrix,
        )
        prev_num_nodes = node_count

    # handles TRs overlapping shards
    i = 0
    # get first subfam in shard
    tr_overlap_index: int = -1
    while i < len(changes_orig) and changes_orig[i] == "skip":
        i += 1

    if i == len(changes_orig):
        # whole shard is skip state
        last_subfam_start = -1
        last_subfam_stop = -1
    else:
        first_subfam_start = (
            non_empty_columns_orig[changes_position_orig[i]] + start_all - 1
        )
        # check if first print pos is one after prev stop in seq pos
        if prev_stop + 1 == first_subfam_start:
            tr_overlap_index = i
            tr_start_seq_pos: int = prev_start - start_all + 1
            stdout.write("\033[2K\033[1G")  # remove last printed line

        # get last subfam in shard
        # will find something since whole shard is not a skip state
        i = len(changes_orig) - 1
        while changes_orig[i] == "skip":
            i -= 1
        last_subfam_start = (
            non_empty_columns_orig[changes_position_orig[i]] + start_all - 1
        )
        last_subfam_stop = (
            non_empty_columns_orig[changes_position_orig[i + 1] - 1]
            + start_all
            - 1
        )

    tr_consensus_changes: Dict[int, str] = {}
    if len(tandem_repeats) > 0:
        # label TRs with their consensus sequence
        changes_index: int = 0
        while changes_index < len(changes_orig):
            subfam: str = changes_orig[changes_index]
            if subfam == "Tandem Repeat":
                cur_changes_pos: int = changes_position_orig[changes_index]

                start_seq_pos: int = (
                    non_empty_columns_orig[cur_changes_pos] + start_all - 1
                )

                if changes_index == tr_overlap_index:
                    # change start pos of first subfam to be prev start seq pos
                    non_empty_columns_orig[cur_changes_pos] = tr_start_seq_pos
                    tr_overlap_index = -1

                stop_seq_pos: int = (
                    non_empty_columns_orig[
                        changes_position_orig[changes_index + 1] - 1
                    ]
                    + start_all
                    - 1
                )

                # track consensus label row
                prev_subfam_row: int = subfam_alignments_collapse[
                    subfam, start_seq_pos
                ][0]
                tr_row: int = prev_subfam_row - (
                    len(alignment_subfamilies) - len(tandem_repeats)
                )
                tr_consensus_changes[changes_index] = (
                    "(" + tr_consensus_matrix[tr_row] + ")n#Simple_repeat"
                )

                # check if TR annotation is from multiple TR rows
                for seq_pos in range(start_seq_pos, stop_seq_pos + 1):
                    cur_subfam_row: int = subfam_alignments_collapse[
                        subfam, seq_pos
                    ][0]
                    if prev_subfam_row != cur_subfam_row:
                        # different TR row used
                        changes_index += 1

                        tr_row = cur_subfam_row - (
                            len(alignment_subfamilies) - len(tandem_repeats)
                        )
                        tr_consensus_changes[changes_index] = (
                            "("
                            + tr_consensus_matrix[tr_row]
                            + ")n#Simple_repeat"
                        )

                        # insert for printing
                        changes_orig.insert(changes_index, "Tandem Repeat")
                        node_ids.insert(
                            changes_index,
                            node_ids[non_empty_columns_orig[cur_changes_pos]],
                        )
                        changes_position_orig.insert(
                            changes_index, cur_changes_pos
                        )
                    cur_changes_pos += 1
                    prev_subfam_row = cur_subfam_row
            changes_index += 1

    # prints results
    if printer.use_matrix_position:
        printer.print_results_matrix(
            changes_orig,
            tr_consensus_changes,
            changes_position_orig,
            non_empty_columns_orig,
            node_ids,
        )
    elif printer.use_sequence_position:
        printer.print_results_sequence(
            start_all,
            changes_orig,
            tr_consensus_changes,
            changes_position_orig,
            non_empty_columns_orig,
            node_ids,
        )
    else:
        printer.print_results_chrom(
            start_all,
            target.chrom_start,
            changes_orig,
            tr_consensus_changes,
            changes_position_orig,
            non_empty_columns_orig,
            node_ids,
        )

    if printer.use_soda_output:
        consensus_lengths = find_consensus_lengths(alignments)
        printer.print_results_soda(
            start_all,
            target.chrom_name,
            target.chrom_start,
            target.chrom_stop,
            alignment_subfamilies,
            changes_orig,
            tr_consensus_changes,
            changes_position_orig,
            non_empty_columns_orig,
            consensus_lengths,
            strand_matrix_collapse,
            consensus_matrix_collapse,
            subfams_collapse_index,
            node_confidence_orig,
            node_ids,
            original_subfamily_sequences,
            original_chromosome_sequences,
            alignment_consensus_start_positions,
            alignment_consensus_stop_positions,
            subfam_alignments_collapse,
            support_matrix_collapse,
            subfams_collapse,
            column_count,
        )

    return last_subfam_start, last_subfam_stop



from typing import Dict, List, Tuple

from polyA.performance import timeit


@timeit()
def fill_path_graph(
    nodes: int,
    columns: List[int],
    changes: List[str],
    changes_position: List[int],
    consensus_matrix_collapse: Dict[Tuple[int, int], int],
    strand_matrix_collapse: Dict[Tuple[int, int], str],
    node_confidence: Dict[Tuple[str, int], float],
    subfams_collapse_index: Dict[str, int],
) -> List[int]:
    """
    finds alternative paths through the nodes - used for stitching to find nested elements
    and to stitch back together elements that have been inserted into.

    Alternative edges only added when on same strand, and confidence of other node label is above a certain
    threshold and when the alignments are relatively contiguous in the consensus sequence.

    input:
    nodes: number of nodes
    columns: list with all non empty columns in matrices
    changes: list of node boundaries
    changes_position: list of subfams identified for each node
    consensus_matrix_collapse: matrix holding alignment position in consensus/subfam seqs -
    used to identify whether nodes are spacially close enough for an alternative edge
    strand_matrix_collapse: matrix holding strand for alignments
    node_confidence: competing annoation confidence for nodes
    subfams_collapse_index: maps subfam names to their row indices in collapsed matrices

    output:
    path_graph: Flat array reprepsentation of 2D matrix. Graph used during stitching. Maps nodes to all other nodes.
    1 if nodes are connected, 0 if not

    >>> non_cols = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> changess = ["s1", "s2", "s1"]
    >>> changes_pos = [0, 3, 6, 10]
    >>> con_mat_col = {(0, 0): 0, (0, 1): 1, (0, 2): 2, (0, 3): 3, (0, 6): 6, (0, 7): 7, (0, 8): 8, (0, 9): 9, (1, 3): 1, (1, 4): 2, (1, 5): 3, (1, 6): 4}
    >>> strand_mat_col = {(0, 0): '+', (0, 1): '+', (0, 2): '+', (0, 3): '+', (0, 6): '+', (0, 7): '+', (0, 8): '+', (0, 9): '+', (1, 3): '-', (1, 4): '-', (1, 5): '-', (1, 6): '-'}
    >>> node_conf = {('s1', 0): 0.9, ('s1', 1): 0.5, ('s1', 2): 0.9, ('s1', 0): 0.1, ('s2', 1): 0.5, ('s2', 2): 0.1}
    >>> sub_col_ind = {'s1': 0, 's2': 1}
    >>> fill_path_graph(3, non_cols, changess, changes_pos, con_mat_col, strand_mat_col, node_conf, sub_col_ind)
    [0, 1, 1, 0, 0, 1, 0, 0, 0]
    """
    path_graph: List[int] = []

    for i in range(nodes * nodes):
        path_graph.append(0)

    # filling beginning path graph with straight line through the nodes
    for i in range(nodes - 1):
        path_graph[i * nodes + i + 1] = 1

    for sink_node_index in range(
        2, nodes
    ):  # don't need to add edges between first 2 nodes because already there
        sink_subfam: str = str(changes[sink_node_index])
        sink_subfam_start: int = consensus_matrix_collapse[
            subfams_collapse_index[sink_subfam],
            columns[changes_position[sink_node_index]],
        ]
        sink_strand: str = strand_matrix_collapse[
            subfams_collapse_index[sink_subfam],
            columns[changes_position[sink_node_index]],
        ]

        if (
            sink_subfam != "skip" and sink_subfam != "Tandem Repeat"
        ):  # don't want to add alternative edges to skip nodes or tandem repeats
            for source_node_index in range(sink_node_index - 1):
                source_subfam: str = changes[source_node_index]
                if source_subfam != "Tandem Repeat":
                    source_conf: float = 0.0
                    sink_conf: float = 0.0
                    if (sink_subfam, source_node_index) in node_confidence:
                        source_conf = node_confidence[
                            sink_subfam, source_node_index
                        ]  # sink subfam confidence in source node
                    if (source_subfam, sink_node_index) in node_confidence:
                        sink_conf = node_confidence[
                            source_subfam, sink_node_index
                        ]  # source subfam confidence in sink node
                    source_subfam_index = subfams_collapse_index[source_subfam]
                    source_col: int = columns[
                        changes_position[source_node_index + 1] - 1
                    ]

                    if (
                        source_subfam_index,
                        source_col,
                    ) in consensus_matrix_collapse:

                        source_subfam_stop = consensus_matrix_collapse[
                            source_subfam_index, source_col
                        ]
                        source_strand = strand_matrix_collapse[
                            source_subfam_index, source_col
                        ]

                        # adds in edge if the subfam of the sink is at the source node and if it's
                        # confidence >= 20%, and if the source is before the sink in the consensus sequence

                        if source_conf >= 0.01 or sink_conf >= 0.01:
                            if (
                                sink_strand == "+"
                                and sink_strand == source_strand
                            ):
                                if source_subfam_stop <= sink_subfam_start + 50:
                                    path_graph[
                                        source_node_index * nodes
                                        + sink_node_index
                                    ] = 1
                            elif (
                                sink_strand == "-"
                                and sink_strand == source_strand
                            ):
                                if source_subfam_stop + 50 >= sink_subfam_start:
                                    path_graph[
                                        source_node_index * nodes
                                        + sink_node_index
                                    ] = 1

    return path_graph



from typing import Dict, TextIO

from .constants import PROB_SKIP, PROB_SKIP_TR
from .exceptions import FileFormatException


def read_prior_counts(
    prior_counts_file: TextIO,
    use_trs: bool,
) -> Dict[str, float]:
    subfam_counts: Dict[str, float] = {"skip": PROB_SKIP}

    if use_trs:
        subfam_counts["Tandem Repeat"] = PROB_SKIP_TR
        prob_skip = PROB_SKIP + PROB_SKIP_TR
    else:
        prob_skip = PROB_SKIP

    # Burn the first line headers
    try:
        next(prior_counts_file)
    except StopIteration:
        raise FileFormatException(prior_counts_file.name, 1)

    total_count: float = 0
    current_line_number = 2

    for line in prior_counts_file:
        try:
            subfam, count = line.strip().split()
            subfam_counts[subfam] = float(count)
            total_count += float(count)
        except ValueError:
            raise FileFormatException(
                prior_counts_file.name,
                current_line_number,
            )

        current_line_number += 1

    for key in subfam_counts:
        subfam_counts[key] = (1 - prob_skip) * subfam_counts[key] / total_count

    return subfam_counts



import re
import logging
from typing import Dict, List, Optional, TextIO, Tuple

from .lambda_provider import LambdaProvider, ConstantLambdaProvider


_logger = logging.root.getChild(__name__)


class ParseError(RuntimeError):
    pass


ScoresDict = Dict[str, int]


class SubMatrix:
    """
    A single substitution matrix, including its name, lambda value, and assumed
    character background frequencies.
    The lambda value may be `None`, in which case it must be computed
    before the matrix can be used.
    The background frequencies may also be 'None' if complexity adjusted scoring
    will not be used.

    A single matrix maps '<char1><char2>' (a pair of characters from
    the input alphabet) to the score from the input substitution
    matrix.

    For example: 'AA' => 8
    """

    lamb: float
    name: str
    scores: ScoresDict
    background_freqs: Optional[Dict[str, float]]

    def __init__(self, name: str, lamb: float):
        self.lamb = lamb
        self.name = name
        self.scores = {}
        self.background_freqs = None


SubMatrixCollection = Dict[str, SubMatrix]


def _parse_matrix_header(line: str) -> Tuple[str, Optional[float]]:
    """
    Attempt to parse a matrix header from the given line, containing
    a matrix name and, optionally, a lambda value.
    """
    clean_line = re.sub(r"^\s+|\s+$", "", line)
    tokens = re.split(r"\s", clean_line)

    if len(tokens) == 0:
        raise ParseError(f"incomplete substitution matrix header: '{line}'")
    if len(tokens) > 2:
        raise ParseError(f"invalid substitution matrix header: '{line}'")

    if len(tokens) == 1:
        return tokens[0], None
    # len(tokens) == 2
    return tokens[0], float(tokens[1])


def _parse_background_freqs(line: str) -> Dict[str, float]:
    """
    Attempt to parse a matrix background frequencies from the given line, containing
    each char and it's assumed background frequency.
    """
    matrix_background_freqs: Dict[str, float] = {}
    if line.strip().upper().startswith("BACKGROUND FREQS"):
        clean_line = re.sub(r"^\s+|\s+$", "", line)
        if line.strip().upper().startswith("BACKGROUND FREQS"):
            # convert to dictionary
            string_dict = clean_line[
                clean_line.find("{") : clean_line.find("}") + 1
            ]
            matrix_background_freqs = eval(string_dict)
    return matrix_background_freqs


def _parse_chars(line: str) -> List[str]:
    clean_line = re.sub(r"^\s+|\s+$", "", line)
    return re.split(r"\s+", clean_line)


def load_substitution_matrices(
    file: TextIO,
    lambda_provider: LambdaProvider,
    complexity_adjustment: bool,
    alphabet_chars: str = "AGCTYRWSKMDVHBXN",
    ambiguity_chars: str = "-.",
) -> SubMatrixCollection:
    """
    Reads a set of score matrices from a file and returns a
    `SubMatrixCollection` that contains the matrices and any lambda
    values associated with them.

    Each matrix has a name, which is then specified for each
    alignment that is to be adjudicated.

    >>> _lamb_provider = ConstantLambdaProvider(0.1227)
    >>> matrices_file = "fixtures/ultra_test_files/ex13.fa.cm.matrix"
    >>> with open(matrices_file) as _sub_matrices_file:
    ...     sub_matrices = load_substitution_matrices(_sub_matrices_file, _lamb_provider, False)
    >>> len(sub_matrices)
    1
    >>> "matrix1" in sub_matrices
    True
    >>> matrix = sub_matrices["matrix1"]
    >>> matrix.lamb
    0.1227
    >>> matrix.name
    'matrix1'
    >>> matrix.scores['AA']
    8
    >>> matrix.scores['AT']
    -15
    >>> matrix.scores['TT']
    8
    >>> matrix.scores['GA']
    -2
    >>> matrix.scores['GC']
    -13
    >>> matrix.scores['CC']
    10
    >>> matrix.background_freqs is None
    True
    >>> with open(matrices_file) as _sub_matrices_file:
    ...     sub_matrices = load_substitution_matrices(_sub_matrices_file, _lamb_provider, True)
    >>> matrix = sub_matrices["matrix1"]
    >>> matrix.background_freqs is None
    False
    >>> matrix.background_freqs
    {'A': 0.295, 'G': 0.205, 'C': 0.205, 'T': 0.295}
    """
    _logger.debug(f"load_substitution_matrix({file.name})")

    collection: SubMatrixCollection = {}

    while True:
        next_scores: ScoresDict = {}
        matrix_background_freqs: Dict[str, float] = {}

        for char1 in alphabet_chars + ambiguity_chars:
            for char2 in alphabet_chars + ambiguity_chars:
                next_scores[char1 + char2] = 0

        try:
            header_line = next(file)
            name, lamb = _parse_matrix_header(header_line)
        except StopIteration:
            break

        try:
            unknown_line = next(
                file
            )  # this could be background freqs or matrix chars
            if unknown_line.strip().upper().startswith("BACKGROUND FREQS"):
                if complexity_adjustment:
                    matrix_background_freqs = _parse_background_freqs(
                        unknown_line
                    )
                    if len(matrix_background_freqs) == 0:
                        # no matrix background frequencies were found
                        raise ParseError(
                            f"cannot use complexity adjusted scoring, missing background frequencies: '{file.name}'"
                        )
                try:
                    # skip background freqs line
                    # this should be the substitution matrix chars
                    unknown_line = next(file)
                except StopIteration:
                    raise ParseError("dangling substitution matrix header")
        except StopIteration:
            break

        chars_line = unknown_line
        chars = _parse_chars(chars_line)

        count: int = 0
        for line in file:
            if "//" in line:
                break

            clean_line = re.sub(r"^\s+|\s+$", "", line)
            sub_scores = re.split(r"\s+", clean_line)
            for i in range(len(sub_scores)):
                next_scores[chars[count] + chars[i]] = int(sub_scores[i])
            count += 1

        if len(next_scores) == 0:
            raise ParseError(
                f"missing substitution matrix values: '{file.name}'"
            )

        if lamb is None:
            lamb = lambda_provider(next_scores)

        sub_matrix = SubMatrix(name, lamb)
        sub_matrix.scores = next_scores
        collection[name] = sub_matrix
        if complexity_adjustment:
            sub_matrix.background_freqs = matrix_background_freqs
    return collection



import os
import re
from tempfile import NamedTemporaryFile
from typing import Callable, Dict

from .performance import timeit

LambdaProvider = Callable[[Dict[str, int]], float]
"""
A callback that accepts a substitution matrix name
and returns the corresponding lambda value for that
matrix.
"""


class ConstantLambdaProvider:
    """
    A test double that returns a fixed value regardless of the
    value passed when called.

    >>> p: LambdaProvider = ConstantLambdaProvider(1.0)
    >>> p({})
    1.0
    """

    _lambda: float

    def __init__(self, value: float):
        self._lambda = value

    def __call__(self, matrix: Dict[str, int]) -> float:
        return self._lambda


class EaselLambdaProvider:
    """
    A provider that executes Easel behind the scenes to recover a
    lambda value for the given matrix.
    """

    _path: str

    def __init__(self, easel_path: str):
        self._path = easel_path

    @timeit()
    def __call__(self, matrix: Dict[str, int]) -> float:
        # Create the temporary matrix file
        with NamedTemporaryFile("w", delete=False) as matrix_file:
            temp_matrix_path = matrix_file.name
            chars = set([k[0] for k in matrix])
            matrix_file.write(" ".join(chars) + "\n")

            for char1 in chars:
                line = " ".join([str(matrix[char1 + char2]) for char2 in chars])
                matrix_file.write(line + "\n")

        # Run Easel
        esl_stream = os.popen(
            self._path + "esl_scorematrix --dna " + temp_matrix_path
        )
        esl_output = esl_stream.read()
        esl_output_list = re.split(r"\n+", esl_output)
        lambda_list = re.split(r"\s+", esl_output_list[1])

        # Clean up by removing the temporary matrix file
        os.remove(temp_matrix_path)

        return float(lambda_list[2])



CHANGE_PROB = 10 ** -55
"""
The base probability of changing annotations.
"""

DEFAULT_CHUNK_SIZE = 31
"""
The width of the "window" used to break a sequence up into segments.
Measured in base pairs.
"""

INFINITE_SHARD_GAP = -1
"""
A magic value that indicates no sharding should occur.
"""

DEFAULT_SHARD_GAP = INFINITE_SHARD_GAP
"""
Allowed gap between sequences for them to be included in the same shard.
"""

NAN_STRING = "NaN"
"""
The string used to represent "not a number" for serialization and
deserialization.
"""

PROB_SKIP = 0.4  # about 60% of genome is TE derived
"""
used for prior counts, about 60% of genome is TE derived, so the skip
state will have a prob of 40%
"""

PROB_SKIP_TR = 0.06
"""
About 6% of the genome is expected to be tandem repeats.
"""

SAME_PROB_LOG = 0.0
"""
penalty for staying int he same state in the DP.
mathematically not exactly 0, but in python log(1-10e-45) = 0
so set to 0 and avoid doing the math
"""

SKIP_ALIGN_SCORE = 10.0
"""
alignment score given to skip state, no lambda adjustment needed
"""

START_ID = 1111
"""
The arbitrary temporary ID used to bootstrap the graph process. The
value doesn't matter, but it's here for clarity.
"""

CROSS_MATCH_ADJUSTMENT = 0.999
"""
A constant value given from cross_match that is used to compute 
the complexity adjusted score of an alignment.
"""

# +-------------------------------------------------------------+
# | Computed constants, do not manually modify below this line. |
# +-------------------------------------------------------------+



from os import path, mkdir
from sys import stdout
from typing import TextIO, Tuple


class Output:
    """
    A class to manage creating output file objects.
    """

    __base_filename: str
    __output_to_file: bool

    def __init__(self, output_path: str, output_to_file: bool):
        base_filename: str
        if path.isdir(output_path):
            base_filename = path.join(output_path, "output")
        else:
            base_filename = output_path

        if base_filename.endswith(path.sep):
            # We got a directory path but it's not an actual
            # directory so we create it then fix the base
            mkdir(base_filename)
            base_filename = path.join(base_filename, "output")

        self.__base_filename = base_filename
        self.__output_to_file = output_to_file

    def get_heatmap(self, index: int) -> TextIO:
        return open(f"{self.__base_filename}.{index}.heatmap", "w")

    def get_results(self) -> TextIO:
        if not self.__output_to_file:
            return stdout

        results_file = open(f"{self.__base_filename}.results", "w")
        return results_file

    def get_soda(self, index: int) -> Tuple[TextIO, TextIO]:
        viz_file = open(f"{self.__base_filename}.{index}.viz", "w")
        conf_file = open(f"{self.__base_filename}.{index}.viz.json", "w")

        return viz_file, conf_file



import sys
import re
from sys import stdout


def read_file(filename_cm):
    """
    opens file and returns contents of file as one string
    """
    file_contents = ''
    with open(filename_cm, 'r') as f_cm:
        file_contents = f_cm.read()

    return file_contents



def get_info(info_array):
    """
    takes info line split by white space and assigns all info fields to variables
    """
    score = info_array[0]
    chrom = info_array[4]
    chrom = info_array[4]

    # doesn't have chrom info because it's an artificial seq
    if 'chr' not in chrom:
        chrom = 'chr0:0000-0000'

    start = info_array[5]
    stop = info_array[6]
    subfam = ''
    consensus_start = ''
    consensus_stop = ''
    strand = ''
    flank = ''

    # if 6 contains (), then start/stop = 7/8
    if "(" in info_array[6]:  # FIXME - shouldn't this be 5?
        start = info_array[6]
        stop = info_array[7]

    #14 for cm .align files, might need to change
    if len(info_array) == 14:  # not compliment
        strand = '+'
        subfam = info_array[8]
        consensus_start = info_array[9]
        consensus_stop = info_array[10]
        flank = info_array[11]

        # if 10 contains (), then start/stop = 7/8
        if "(" in info_array[9]:
            flank = info_array[9]
            start = info_array[10]
            stop = info_array[11]

    else:  # compliment
        strand = '-'
        subfam = info_array[9]
        consensus_start = info_array[10]
        consensus_stop = info_array[11]
        flank = info_array[12]

        if "(" in info_array[10]:
            flank = info_array[10]
            consensus_start = info_array[11]
            consensus_stop = info_array[12]

    flank = flank.strip(")")
    flank = flank.strip("(")

    return subfam, chrom, score, strand, start, stop, consensus_start, consensus_stop, flank


def get_score_matrix(matrix_name):
    """
    grabs score matrix info from original matrix file, puts it in correct format, returns string
    """
    matrix_file_contents = read_file("../fixtures/matrices/" + matrix_name)
    matrix_file_contents = matrix_file_contents.split('\n')
    background_freqs_line = matrix_file_contents[0]
    # reformat background freqs
    background_freqs = background_freqs_line.split()[1:]
    background_freqs_dict = {}
    for i in range(0, len(background_freqs), 2):
        char = background_freqs[i][0]
        freq = float(background_freqs[i + 1])
        background_freqs_dict[char] = freq
    freqs_string = "BACKGROUND FREQS: "+ str(background_freqs_dict)
    return "\n".join(([freqs_string] + matrix_file_contents[1:]))


def print_info(C, subfam, chrom, score, strand, start, stop, consensus_start, consensus_stop, flank, matrix_name, f_out_sto):
    """
    prints all info in correct format for stockholm
    """
    f_out_sto.write(f'#=GF ID  {subfam}\n')
    f_out_sto.write(f'#=GF TR  {chrom}\n')
    f_out_sto.write(f'#=GF SC  {score}\n')
    f_out_sto.write(f'#=GF SD  {strand}\n')

    if strand == '-':
        if C:
            f_out_sto.write(f'#=GF TQ  t\n')
        else:
            f_out_sto.write(f'#=GF TQ  q\n')
    else:
        f_out_sto.write(f'#=GF TQ  -1\n')

    f_out_sto.write(f'#=GF ST  {start}\n')
    f_out_sto.write(f'#=GF SP  {stop}\n')
    f_out_sto.write(f'#=GF CST {consensus_start}\n')
    f_out_sto.write(f'#=GF CSP {consensus_stop}\n')
    f_out_sto.write(f'#=GF FL  {flank}\n')
    f_out_sto.write(f'#=GF MX  {matrix_name}\n')


def get_alignment(alignment_array):
    """
    takes cm alignment split on newline and return string of chrom seq and subfam seq
    """
    chrom_seq = ""
    subfam_seq = ""

    # for (my $i = 0; $i < @ Alignment; $i = $i + 4){
    for i in range(0, len(alignment_array), 4):
        m_chrom = re.search(r'.+?\s\d+\s(.+?)\s\d+', alignment_array[i])
        if m_chrom:
            chrom_seq += m_chrom.group(1)

        m_subfam = re.search(r'.+?\s\d+\s(.+?)\s\d+', alignment_array[i+2])
        if m_chrom:
            subfam_seq += m_subfam.group(1)

    return chrom_seq, subfam_seq


def print_alignment(chrom_seq, subfam_seq, chrom, subfam, f_out_sto):
    """
    takes chrom and subfam names and chrom and subfam seqs and prints in
    stockholm format
    """
    f_out_sto.write(f"{chrom}   {chrom_seq}\n")
    f_out_sto.write(f"{subfam}   {subfam_seq}\n")

    f_out_sto.write("//\n")


def print_score_matrix(f_out_matrix, score_matrix, matrix_name):
    """
    print score matrix to its output file with extension ".matrix"
    """
    f_out_matrix.write(matrix_name)
    f_out_matrix.write("\n")
    f_out_matrix.write(score_matrix)
    f_out_matrix.write("//\n")


if __name__ == "__main__":
    matrices = {}
    filename_rm = sys.argv[1]
    file_contents = read_file(filename_rm)

    filename_out_sto = filename_rm + ".sto"
    f_out_sto = open(filename_out_sto, 'w')

    filename_out_matrix = filename_rm + ".matrix"
    f_out_matrix = open(filename_out_matrix, 'w')

    f_out_sto.write("# STOCKHOLM 1.0\n")
    f_out_sto.write(f"# ALIGNMENT TOOL RepeatMasker\n")

    alignments = re.findall(r'\s*?\d+\s+[0-9]+\.[0-9]+\s+[0-9.]+\s+[0-9.]+\s+.+?\n\n[\s\S]+?Transitions', file_contents)

    for region in alignments:
        info_line: str = ''
        alignment: str = ''
        matrix_name: str = ''
        region = region.strip()

        m_matrix = re.search(r'Matrix = (.+?)\n', region)
        matrix_name = m_matrix[1]

        if matrix_name not in matrices:  # do not put duplicate of matrices in output file
            matrices[matrix_name] = 0
            score_matrix = get_score_matrix(matrix_name)
            print_score_matrix(filename_out_matrix, score_matrix, matrix_name)

        m = re.match(r'(.+?)\n([\s\S]+)\n\nMatrix', region)
        if m:
            info_line = m.group(1)
            alignment = m.group(2)

        alignment = alignment.strip("\n")

        info_array = info_line.split()

        subfam, chrom, score, strand, start, stop, consensus_start, consensus_stop, flank = get_info(info_array)

        alignment_array = alignment.split("\n")

        # if C is true: on reverse strand and target is reversed
        C = 0
        if "C" in alignment_array[0]:
            C = 1

        chrom_seq, subfam_seq = get_alignment(alignment_array)

        print_info(C, subfam, chrom, score, strand, start, stop, consensus_start, consensus_stop, flank, matrix_name, f_out_sto)
        print_alignment(chrom_seq, subfam_seq, chrom, subfam, f_out_sto)

    f_out_sto.close()



from typing import Dict, List, Tuple

from .confidence_cm import confidence_cm
from .matrices import ConfidenceMatrix
from .performance import timeit


@timeit()
def fill_confidence_matrix(
    column_count: int,
    subfam_counts: Dict[str, float],
    subfams: List[str],
    active_cells: Dict[int, List[int]],
    align_matrix: Dict[Tuple[int, int], float],
) -> ConfidenceMatrix:
    """
    Fills confidence matrix from alignment matrix. Each column in the alignment
    matrix is a group of competing annotations that are input into
    confidence_cm, the output confidence values are used to populate
    confidence_matrix.

    Inputs:

    columns - non empty matrix column indices
    subfam_counts - mapping of subfamily names to their prior counts
    subfams - list of subfamily names taken from the original alignments
    active_cells - map of column indices (from columns) into list of non-empty
    rows for the given column
    align_matrix - the alignment matrix from fill_align_matrix

    Outputs:

    confidence_matrix - hash implementation of sparse 2D matrix used in pre-DP
    calculations. Key is tuple[int, int] that maps row, col with the value held
    in that cell of matrix. Rows are subfamilies in the input alignment file,
    cols are nucleotide positions in the alignment. Each cell in matrix is the
    confidence score calculated from all the alignment scores in a column of the
    alignment matrix.

    >>> align_mat = {
    ...     (0, 0): 10, (0, 1): 10, (0, 2): 10, (0, 3): 10, (0, 4): 10,
    ...     (1, 1): 42, (1, 2): 41, (1, 3): 41, (2, 1): 45, (2, 2): 43, (2, 3): 39,
    ... }
    >>> active = {0: [0], 1: [0, 1, 2], 2: [0, 1, 2], 3: [0, 1, 2], 4: [0]}
    >>> col_count = 5
    >>> counts = {"skip": 0.4, "s1": .33, "s2": .33}
    >>> subs = ["skip", "s1", "s2"]
    >>> conf_mat = fill_confidence_matrix(col_count, counts, subs, active, align_mat)
    >>> f"{conf_mat[1,1]:.4f}"
    '0.1100'
    >>> f"{conf_mat[2,1]:.4f}"
    '0.8800'
    >>> f"{conf_mat[1,2]:.4f}"
    '0.1980'
    >>> f"{conf_mat[2,2]:.4f}"
    '0.7920'
    >>> f"{conf_mat[1,3]:.4f}"
    '0.7920'
    >>> f"{conf_mat[2,3]:.4f}"
    '0.1980'
    """
    confidence_matrix: ConfidenceMatrix = {}

    for col_index in range(column_count):
        temp_region: List[float] = []

        for row_index in active_cells[col_index]:
            temp_region.append(align_matrix[row_index, col_index])

        temp_confidence: List[float] = confidence_cm(
            temp_region,
            subfam_counts,
            subfams,
            active_cells[col_index],
            0,
            False,
        )

        for row_index2 in range(len(active_cells[col_index])):
            confidence_matrix[
                active_cells[col_index][row_index2], col_index
            ] = temp_confidence[row_index2]

    return confidence_matrix



from polyA.alignment import Alignment, get_skip_state
from polyA.load_alignments import load_alignments, shard_overlapping_alignments


def test_load_alignments_1():
    with open(f"fixtures/alignment1.sto", "r") as file:
        skip, first, second = load_alignments(file, add_skip_state=True)

    assert first.subfamily == "L2#LINE/L2"
    assert second.subfamily == "Charlie12#DNA/hAT-Charlie"

    assert first.sequence.startswith("ATTTCCAG---")
    assert second.sequence.startswith("AAGCTTGTCCAA")

    assert first.subfamily_sequence.startswith("ATCTCCAGCCCA")
    assert second.subfamily_sequence.startswith("AAGCTTGTCCA")

    assert first.score == 245
    assert second.score == 671

    assert first.start == 323
    assert second.start == 983

    assert first.stop == 499
    assert second.stop == 1095

    assert first.consensus_start == 2349
    assert second.consensus_start == 2

    assert first.consensus_stop == 2534
    assert second.consensus_stop == 118

    assert first.flank == 548
    assert second.flank == 2755

    assert first.sub_matrix_name == "matrix1"
    assert second.sub_matrix_name == "matrix2"

    assert first.gap_init == 1
    assert second.gap_init == 2

    assert first.gap_ext == 11
    assert second.gap_ext == 22


# test to make sure seqs get flipped with reverse is on target (TQ == 't')
def test_load_alignments_2():
    with open(f"fixtures/alignment2.sto", "r") as file:
        skip, first = load_alignments(file, add_skip_state=True)

    assert first.sequence.startswith("AACAAGAA")
    assert first.subfamily_sequence.startswith("AAAAAAAAA")


def test_load_alignments_no_skip_state():
    with open(f"fixtures/alignment2.sto", "r") as file:
        alignments = list(load_alignments(file, add_skip_state=False))

    assert alignments[0] is not get_skip_state()


def test_shard_alignments_simple():
    # The first shard looks like this:
    #   1   5    10   15   20
    # 0 |--------|
    # 1     |---------|
    # 2  |------|
    s00 = Alignment("", "a", 1, 1000, 0, 1, 10, 0, 0, [], "", 0, "", 0, 0)
    s01 = Alignment("", "a", 1, 1000, 0, 5, 15, 0, 0, [], "", 0, "", 0, 0)
    s02 = Alignment("", "a", 1, 1000, 0, 2, 8, 0, 0, [], "", 0, "", 0, 0)

    # The second shard looks the same, but it is shifted
    # to the right by 115 positions (100 positions beyond
    # right-most position included in the first shard).
    offset = 100 + 15
    s10 = Alignment(
        "",
        "a",
        1,
        1000,
        0,
        s00.start + offset,
        s00.stop + offset,
        0,
        0,
        [],
        "",
        0,
        "",
        0,
        0,
    )
    s11 = Alignment(
        "",
        "a",
        1,
        1000,
        0,
        s01.start + offset,
        s01.stop + offset,
        0,
        0,
        [],
        "",
        0,
        "",
        0,
        0,
    )
    s12 = Alignment(
        "",
        "a",
        1,
        1000,
        0,
        s02.start + offset,
        s02.stop + offset,
        0,
        0,
        [],
        "",
        0,
        "",
        0,
        0,
    )

    skip = get_skip_state()

    shards = list(
        shard_overlapping_alignments([s00, s01, s02, s10, s11, s12], 50)
    )

    assert len(shards) == 2

    assert shards[0].start == 1
    assert shards[0].stop == 65
    assert shards[0].alignments == [skip, s00, s01, s02]

    assert shards[1].start == 66
    assert shards[1].stop == 1000
    assert shards[1].alignments == [skip, s10, s11, s12]


def test_shard_alignments_fixture():
    with open(f"fixtures/ultra_test_files/ex27.fa.cm.sto") as file:
        alignments = load_alignments(file, add_skip_state=False)
        shards = list(shard_overlapping_alignments(alignments, 50, False))

    shard0 = shards[0]
    assert len(shard0.alignments) == 12

    shard1 = shards[1]
    assert len(shard1.alignments) == 27

    shard_last = shards[-1]
    assert shard_last.stop == 146336591 - 146331457 + 1
